﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Threading.Tasks;
using Bots.DungeonBuddy.Attributes;
using Bots.DungeonBuddy.Enums;
using Bots.DungeonBuddy.Helpers;
using Bots.DungeonBuddy.Profiles;
using Buddy.Coroutines;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Bars;
using Styx.CommonBot.Coroutines;
using Styx.CommonBot.Routines;
using Styx.Helpers;
using Styx.Pathing;
using Styx.Pathing.Avoidance;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

// ReSharper disable CheckNamespace

namespace Bots.DungeonBuddy.DungeonScripts.Legion
// ReSharper restore CheckNamespace
{
    public abstract class TheNighthold : Dungeon
    {
        #region Overrides of Dungeon
        public override Vector3 Entrance { get; } = new Vector3(1237.402f, 4205.289f, -65.37196f);

        public override Vector3 ExitLocation { get; } = new Vector3(-157.675f, 3514.278f, -252.5676f);


        public override void OnEnter()
        {
            if (Me.IsTank())
            {
                Alert.Show(
                    "Tanking Not Supported",
                    string.Format(
                        "Tanking is not supported in the {0} script. " +
                        "If you wish to stay in raid and play manually then press 'Continue', otherwise you will automatically leave raid.",
                        Name),
                    30,
                    true,
                    true,
                    null,
                    () => Lua.DoString("LeaveParty()"),
                    "Continue",
                    "Leave");
            }
            else
            {
                Alert.Show(
                    "Do Not AFK",
                    "It is highly recommended you do not afk while in a raid and be prepared to intervene " +
                    "if needed in the event something goes wrong or you're asked to perform a certain task.",
                    20,
                    true,
                    false,
                    null,
                    null,
                    "Ok");
            }
        }

        #endregion
        protected static LocalPlayer Me => StyxWoW.Me;
    }

    public class ArcingAqueducts : TheNighthold
    {
        #region Overrides of Dungeon

        public override uint DungeonId => 1290;

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            units.RemoveAll(
                ret =>
                {
                    var unit = ret as WoWUnit;
                    if (unit == null)
                        return false;

                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            var isTank = Me.IsTank;
            var inDungeon = Me.MapId == LfgDungeon.MapId;

            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit == null)
                    continue;
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            var isDps = Me.IsDps;

            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit == null)
                    continue;

                switch (unit.Entry)
                {
                    case MobId_WaningTimeParticle:
                        priority.Score += 5500;
                        break;

                    case MobId_FragmentedTimeParticle:
                        priority.Score += 5000;
                        break;

                    case MobId_Scrubber:
                        priority.Score += unit.CastingSpellId == SpellId_Scrubbing && Me.HasAura("Succulent Feast") ? 5000 : -5000;
                        break;
                }
            }
        }


        #endregion

        #region Root


        [EncounterHandler(0, "Root Handler")]
        public Func<WoWUnit, Task<bool>> RootBehavior()
        {
            CapabilityManagerHandle handle = CapabilityManager.Instance.CreateNewHandle();
            return async npc =>
            {
                return false;
            };
        }

        #endregion

        #region Skorpyron

        private const int SpellId_Arcanoslash = 204275;
        private const int SpellId_FocusedBlast = 204471;

        private const uint MobId_CrystallineShard = 103209;

        private const uint MobId_Skorpyron = 102263;
        private const int SpellId_Shockwave = 204316;

        [EncounterHandler(MobId_Skorpyron, "Skorpyron")]
        public Func<WoWUnit, Task<bool>> SkorpyronHandler()
        {
            Func<WoWUnit, WoWUnit, Vector3> calculateSpikeLosLoc
                = (boss, spike) => WoWMathHelper.CalculatePointFrom(boss.Location, spike.Location, -5);

            WoWUnit skorpyron = null;
            AddAvoidObject<WoWUnit>(20,
                u => u.Entry == MobId_Skorpyron && u.Combat,
                u => u.Location.RayCast(u.Rotation, 18));

            Vector2[] focusedBlastPoly = new[] { new Vector2(0, 0), new Vector2(3, -8), new Vector2(10000, 0), new Vector2(3, 8) };

            AddAvoidPolygon(() => ScriptHelpers.IsViable(skorpyron) && skorpyron.CastingSpellId == SpellId_FocusedBlast,
                null, 40, u => u.Rotation, u => 1, u => 40,
                u => focusedBlastPoly, u => u.Location, () => new[] { skorpyron });

            return async boss =>
            {
                skorpyron = boss;

                if (boss.CastingSpellId == SpellId_Shockwave)
                {
                    var spikeLosLoc = (from unit in ObjectManager.GetObjectsOfType<WoWUnit>()
                                       where unit.Entry == MobId_CrystallineShard
                                       let losLoc = calculateSpikeLosLoc(boss, unit)
                                       orderby Me.Location.DistanceSquared(losLoc)
                                       select losLoc).FirstOrDefault();

                    if (spikeLosLoc != Vector3.Zero)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => ScriptHelpers.IsViable(boss) && boss.CastingSpellId == SpellId_Shockwave && spikeLosLoc != Vector3.Zero,
                            spikeLosLoc, "Shockwave", 1);
                    }
                }
                return false;
            };
        }

        #endregion

        #region Chronomatic Anomaly

        private const uint MobId_ChronomaticAnomaly = 104415;
        private const int MissileSpellId_BurstofTime = 206613;
        private const uint AreaTriggerId_TemporalOrb = 11521;

        [EncounterHandler(MobId_ChronomaticAnomaly, "Chronomatic Anomaly")]
        public Func<WoWUnit, Task<bool>> ChronomaticAnomalyHandler()
        {

            AddAvoidLocation(() => true, m => 6, m => m.ImpactPosition,
                () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == MissileSpellId_BurstofTime));

            AddAvoidObject<WoWAreaTrigger>(1.75f, AreaTriggerId_TemporalOrb);

            return async boss =>
            {
                if (Me.IsDps && Me.GotTarget && Me.CurrentTarget.Entry != MobId_WaningTimeParticle)
                {
                    var waningTimeParticle = ObjectManager.GetObjectsOfType<WoWUnit>()
                        .Where(u => u.Entry == MobId_WaningTimeParticle && (!u.Attackable || !u.CanSelect))
                        .MinByOrDefault(u => u.DistanceSqr);

                    // Move nearby to quickly dps when Waning Time Particle becomes attackable.
                    if (waningTimeParticle != null)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => ScriptHelpers.IsViable(waningTimeParticle) && (!waningTimeParticle.Attackable || !waningTimeParticle.CanSelect),
                            waningTimeParticle.Location, waningTimeParticle.SafeName, Me.IsRanged ? 35 : waningTimeParticle.MeleeRange);
                    }
                }
                return false;
            };
        }

        private const int SpellId_WarpNightwell = 207228;
        private const uint MobId_WaningTimeParticle = 104676;

        [EncounterHandler(MobId_WaningTimeParticle, "Waning Time Particle")]
        public Func<WoWUnit, Task<bool>> WaningTimeParticleHandler()
        {
            return async boss => await ScriptHelpers.InterruptCast(boss, SpellId_WarpNightwell);
        }

        private const int SpellId_WarpNightwell2 = 228335;
        private const uint MobId_FragmentedTimeParticle = 114671;
        [EncounterHandler(MobId_FragmentedTimeParticle, "Fragmented Time Particle")]
        public Func<WoWUnit, Task<bool>> FragmentedTimeParticleHandler()
        {
            return async boss => await ScriptHelpers.InterruptCast(boss, SpellId_WarpNightwell);
        }

        #endregion

        #region Trilliax

        private const uint AreaTriggerId_ArcaneSeepage = 10928;
        private const uint AreaTriggerId_ToxicSlice = 10958;
        private const uint AreaTriggerId_SlimePool2 = 12514;
        private const uint AreaTriggerId_SlimePool = 12513;
        private const uint AreaTriggerId_BurstingSlime = 12531;

        private const int SpellId_Annihilation = 207630;
        private const int SpellId_Scrubbing = 211907;
        private const uint MobId_SucculentFeast = 104561;
        private const uint MobId_Scrubber = 104596;
        private const uint MobId_Trilliax = 104288;

        [EncounterHandler(MobId_Trilliax, "Trilliax")]
        public Func<WoWUnit, Task<bool>> TrilliaxHandler()
        {
            WoWUnit trilliax = null;
            AddAvoidObject<WoWAreaTrigger>(4, AreaTriggerId_ArcaneSeepage, AreaTriggerId_BurstingSlime);
            AddAvoidObject<WoWAreaTrigger>(5, AreaTriggerId_SlimePool, AreaTriggerId_SlimePool2);

            AddAvoidObject<WoWAreaTrigger>(() => ScriptHelpers.IsViable(trilliax)
                && trilliax.Combat
                && (!trilliax.HasAura("The Cleaner") || !Me.HasAura("Toxic Slice")),
                4, a => a.Entry == AreaTriggerId_ToxicSlice);

            AddAvoidLine<WoWUnit>(() => true, u => WoWMathHelper.CalculatePointBehind(u.Location, u.Rotation, 100),
                u => u.Rotation, u => 200, u => 5,
                u => u.Entry == MobId_Trilliax && u.CastingSpellId == SpellId_Annihilation);
            AddAvoidObject<WoWPlayer>(() => true, 7, p => !p.IsMe && (p.HasAura("Sterilize") || Me.HasAura("Sterilize")));

            return async boss =>
            {
                trilliax = boss;

                if (boss.HasAura("The Cleaner") && !Me.HasAura("Toxic Slice"))
                {
                    var toxicSlice = ObjectManager.GetObjectsOfType<WoWAreaTrigger>()
                        .Where(a => a.Entry == AreaTriggerId_ToxicSlice)
                        .MinByOrDefault(a => a.DistanceSqr);
                    if (toxicSlice != null)
                        return (await CommonCoroutines.MoveTo(toxicSlice.Location, toxicSlice.SafeName)).IsSuccessful();
                }

                // Not sure if this ability is used in LFR difficulty
                if (Me.HasAura("Arcing Bonds"))
                {
                    var match = Me.RaidMembers.FirstOrDefault(r => !r.IsMe && r.HasAura("Arcing Bonds"));
                    if (match != null)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => Me.HasAura("Arcing Bonds") && ScriptHelpers.IsViable(match),
                            match.Location, "Arcing Bonds meetup");
                    }
                }

                if (Me.IsDps && !Me.HasAura("Succulent Feast"))
                {
                    var feast = ObjectManager.GetObjectsOfType<WoWUnit>()
                        .Where(u => u.Entry == MobId_SucculentFeast)
                        .MinByOrDefault(u => u.DistanceSqr);

                    if (feast != null)
                        return await ScriptHelpers.InteractWithObject(feast, 100, true);
                }

                if (Me.HasAura("Sterilize"))
                {
                    var scrubber = ObjectManager.GetObjectsOfType<WoWUnit>()
                       .Where(u =>
                       {
                           if (u.Entry != MobId_Scrubber || Blacklist.Contains(u, BlacklistFlags.Interact))
                               return false;

                           if (u.ManaPercent < 10)
                           {
                               Blacklist.Add(u, BlacklistFlags.Interact, TimeSpan.FromSeconds(30));
                               return false;
                           }
                           return true;
                       })
                       .MinByOrDefault(u => u.DistanceSqr);

                    if (scrubber != null)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => ScriptHelpers.IsViable(scrubber) && Me.HasAura("Sterilize"), 
                            scrubber.Location, scrubber.SafeName, 6);
                    }
                }

                return false;
            };
        }

        #endregion

    }
}