﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Threading.Tasks;
using Bots.DungeonBuddy.Attributes;
using Bots.DungeonBuddy.Enums;
using Bots.DungeonBuddy.Helpers;
using Bots.DungeonBuddy.Profiles;
using Buddy.Coroutines;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Bars;
using Styx.CommonBot.Coroutines;
using Styx.CommonBot.Routines;
using Styx.Helpers;
using Styx.Pathing;
using Styx.Pathing.Avoidance;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Styx.WoWInternals.WoWObjects.AreaTriggerShapes;

// ReSharper disable CheckNamespace

namespace Bots.DungeonBuddy.DungeonScripts.Legion
// ReSharper restore CheckNamespace
{
    public class BetrayersRise : TheNighthold
    {
        #region Overrides of Dungeon

        public override uint DungeonId => 1293;

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            units.RemoveAll(
                ret =>
                {
                    var unit = ret as WoWUnit;
                    if (unit == null)
                        return false;

                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            var isTank = Me.IsTank;
            var inDungeon = Me.MapId == LfgDungeon.MapId;

            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit == null)
                    continue;
            }
        }
        
        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            var isDps = Me.IsDps;

            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit == null)
                    continue;

                switch (unit.Entry)
                {
                    case MobId_EyeofGuldan:
                    case MobId_EmpoweredEyeofGuldan:
                        priority.Score += 5500;
                        break;
                }
            }
        }

        public override void IncludeLootTargetsFilter(List<WoWObject> incomingObjects, HashSet<WoWObject> outgoingObjects)
        {
            foreach (var lootTarget in incomingObjects)
            {
                var gObj = lootTarget as WoWGameObject;
                if (gObj != null && gObj.Entry == GameObjectId_GiftoftheNightborne && gObj.CanLoot &&
                    DungeonBuddySettings.Instance.LootMode != LootMode.Off)
                {
                    outgoingObjects.Add(gObj);
                }
            }
        }

        public override void OnEnter()
        {
            base.OnEnter();
            _capabilityHandle = CapabilityManager.Instance.CreateNewHandle();
        }

        public override void OnExit()
        {
            base.OnExit();
            if (_capabilityHandle != null)
            {
                CapabilityManager.Instance.Clear(_capabilityHandle);
                _capabilityHandle = null;
            }
        }

        #endregion

        #region Root

        CapabilityManagerHandle _capabilityHandle;

        [EncounterHandler(0, "Root Handler")]
        public Func<WoWUnit, Task<bool>> RootBehavior()
        {
            return async npc => await ScriptHelpers.CancelCinematicIfPlaying();
        }

        #endregion

        #region Gul'dan

        private const uint MobId_Guldan = 104154;
        private const uint MobId_InquisitorVethriz = 104536;
        private const uint MobId_EmpoweredEyeofGuldan = 106545;
        private const uint MobId_EyeofGuldan = 105630;
        private const uint MobId_LiquidHellfire = 104396;

        private const uint AreaTriggerId_FelEfflux = 9575;
        private const uint AreaTriggerId_SoulVortex = 10971;
        private const uint AreaTriggerId_ScatteringField = 12015;
        private const uint AreaTriggerId_EmpoweredLiquidHellfire = 10936;
        private const uint AreaTriggerId_FlamesofSargeras = 12376;
        private const uint AreaTriggerId_WellofSouls = 9660;
        private const uint AreaTriggerId_GazeofVethriz = 12001;

        private const int MissileSpellId_HandofGuldan = 212258;
        private const int MissileSpellId_LiquidHellfire = 206554;
        private const int MissileSpellId_EmpoweredLiquidHellfire = 206586;
        private const uint GameObjectId_GiftoftheNightborne = 266483;
        private const uint GameObjectId_EyeofAmunthul = 250082;

        private const int SpellId_EyeofGuldan = 209454;
        private const int SpellId_ScatteringField = 217830;
        private const int SpellId_TimeDilation = 210339;

        private const int SpellId_StormoftheDestroyer_10y = 167819;
        private const int SpellId_StormoftheDestroyer_20y = 167935;
        private const int SpellId_StormoftheDestroyer_30y = 177380;
        private const int SpellId_StormoftheDestroyer_60y = 152987;

        [ObjectHandler((int)GameObjectId_GiftoftheNightborne, ObjectRange = 100)]
        public async Task<bool> GiftoftheNightborneHandler(WoWGameObject chest)
        {
            if (ScriptHelpers.IsBossAlive("Gul'dan"))
                ScriptHelpers.MarkBossAsDead("Gul'dan", "Chest has spawned.");
            return false;
        }

        [EncounterHandler(MobId_Guldan, "Gul'dan")]
        public Func<WoWUnit, Task<bool>> GuldanHandler()
        {
            WoWUnit guldan = null;
            AddAvoidObject<WoWUnit>(() => true, 25, u => u.Entry == MobId_Guldan && u.CastingSpellId == SpellId_StormoftheDestroyer_10y);
            AddAvoidObject<WoWUnit>(() => true, 40, u => u.Entry == MobId_Guldan && u.CastingSpellId == SpellId_StormoftheDestroyer_20y);
            AddAvoidObject<WoWUnit>(() => true, 55, u => u.Entry == MobId_Guldan && u.CastingSpellId == SpellId_StormoftheDestroyer_30y);
            AddAvoidObject<WoWUnit>(() => true, 60, u => u.Entry == MobId_Guldan && u.CastingSpellId == SpellId_StormoftheDestroyer_60y);
            AddAvoidObject<WoWGameObject>(14, GameObjectId_EyeofAmunthul);
            AddAvoidObject<WoWGameObject>(15, AreaTriggerId_GazeofVethriz);

            // Stand away from Cul'dan's current target to prevent getting cleaved
            AddAvoidObject<WoWUnit>(() => true, 2, 
                u => u.Entry == MobId_Guldan && u.HasAura("Fel Scythe") && u.GotTarget, 
                u => u.CurrentTarget.Location);

            var fellEffluxPoly = new []{ new Vector2(0, 6.5f),new Vector2(0, -6.5f),new Vector2(100, -6.5f),new Vector2(100, 6.5f)};
            AddAvoidPolygon(() => true, null, 40, a => a.Rotation, a => 1, a => 40, a => fellEffluxPoly, a => a.Location, 
                () => ObjectManager.GetObjectsOfType<WoWAreaTrigger>().Where(a => a.Entry == AreaTriggerId_FelEfflux),
                a => a.IsValid);

            AddAvoidLocation(() => true, 10, m => m.ImpactPosition,
                () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == MissileSpellId_HandofGuldan));

            AddAvoidLocation(() => true, 10, m => m.ImpactPosition,
                () => WoWMissile.InFlightMissiles
                    .Where(m => m.SpellId == MissileSpellId_LiquidHellfire
                             || m.SpellId == MissileSpellId_EmpoweredLiquidHellfire));

            AddAvoidObject<WoWAreaTrigger>(5, AreaTriggerId_SoulVortex, AreaTriggerId_FlamesofSargeras);
            AddAvoidObject<WoWAreaTrigger>(()=> true, a => ((AreaTriggerSphere)a.Shape).CurrentRadius + 1, AreaTriggerId_EmpoweredLiquidHellfire);

            Func<WoWUnit, bool> hasEyeOfGuldan = player => player.HasAura("Eye of Gul'dan") || player.HasAura("Empowered Eye of Gul'dan");
            AddAvoidObject<WoWPlayer>(() => true, a => 8, p => !p.IsMe && (hasEyeOfGuldan(p) || hasEyeOfGuldan(Me)));
            AddAvoidObject<WoWPlayer>(() => true, a => Me.HasAura("Flames of Sargeras") ? 20: 8,
                p => !p.IsMe && (p.HasAura("Flames of Sargeras") || Me.HasAura("Flames of Sargeras")));

            AddAvoidObject<WoWAreaTrigger>(() => true, 14, a => a.Entry == AreaTriggerId_WellofSouls);

            TimeCachedValue<Vector3> groupLocation = new TimeCachedValue<Vector3>(3000, () => ScriptHelpers.GetGroupCenterLocation(null, 5));
            return async boss =>
            {
                guldan = boss;
                if (Me.HasAura("Bonds of Fel") || Me.HasAura("Empowered Bonds of Fel"))
                {
                    var raidMembersNearby = Me.RaidMembers.Count(r => !r.IsMe && r.Location.DistanceSquared(Me.Location) < 5 * 5 && r.IsAlive);
                    Logger.WriteDebug($"Raidmembers Nearby: {raidMembersNearby}");
                    if (raidMembersNearby > 0)
                    {
                        var moveto = Me.RaidMembers.Where(r => !r.IsMe && r.IsAlive && r.DistanceSqr > 15 * 15
                                && !AvoidanceManager.Avoids.Any(a => a.IsPointInAvoid(r.Location)))
                            .Select(r => r.Location)
                            .MinByOrDefault(r => r.DistanceSquared(Me.Location));

                        return (await CommonCoroutines.MoveTo(moveto, "Breaking Bonds of Fel")).IsSuccessful();
                    }

                    return await ScriptHelpers.StayAtLocationWhile(
                        () => (Me.HasAura("Bonds of Fel") || Me.HasAura("Empowered Bonds of Fel")) 
                            && Me.RaidMembers.Count(r => !r.IsMe && r.Location.DistanceSquared(Me.Location) < 5*5 && r.IsAlive) == 0, 
                        Me.Location, "Bonds of Fel", 1);
                }
                else if (Me.IsRangedDps)
                {
                    var bondsOfFellTarget = Me.RaidMembers
                        .Where(r => !r.IsMe && (r.HasAura("Bonds of Fel") || r.HasAura("Empowered Bonds of Fel")) 
                            && !Me.RaidMembers.Any(g => !g.IsMe && g != r && g.Location.DistanceSquared(r.Location) < 5*5 ))
                        .MinByOrDefault(r => r.DistanceSqr);

                    if (bondsOfFellTarget != null)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => ScriptHelpers.IsViable(bondsOfFellTarget)
                                && (bondsOfFellTarget.HasAura("Bonds of Fel") || bondsOfFellTarget.HasAura("Empowered Bonds of Fel")),
                            bondsOfFellTarget.Location, "Bonds of Fel Assist", 2);
                    }
                }
                var liquidHellfire = ObjectManager.GetObjectsOfType<WoWUnit>().Any(u => u.Entry == MobId_LiquidHellfire) 
                    || WoWMissile.InFlightMissiles.Any(m => m.SpellId == MissileSpellId_LiquidHellfire 
                                                    || m.SpellId == MissileSpellId_EmpoweredLiquidHellfire);

                if (Me.IsRanged && liquidHellfire)
                {
                    if (Me.IsHealer && ActionBar.Extra.IsActive && ActionBar.Extra.Buttons[0].CanUse)
                    {
                        ActionBar.Extra.Buttons[0].Use();
                        await Coroutine.Wait(10000, () => Me.IsAlive);
                    }

                    var scatteringField = ObjectManager.GetObjectsOfType<WoWAreaTrigger>()
                        .Where(a => a.Entry == AreaTriggerId_ScatteringField)
                        .MinByOrDefault(a => a.Location.DistanceSquared(boss.Location));

                    if (scatteringField != null)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => ScriptHelpers.IsViable(scatteringField),
                            scatteringField.Location, "Scattering Field");
                    }
                }

                if (Me.IsDps && ActionBar.Extra.IsActive && ActionBar.Extra.Buttons[0].CanUse)
                {
                    ActionBar.Extra.Buttons[0].Use();
                    return true;
                }

                return false;
            };
        }

        #endregion
    }
}