﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Threading.Tasks;
using Bots.DungeonBuddy.Attributes;
using Bots.DungeonBuddy.Enums;
using Bots.DungeonBuddy.Helpers;
using Bots.DungeonBuddy.Profiles;
using Buddy.Coroutines;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Bars;
using Styx.CommonBot.Coroutines;
using Styx.CommonBot.Routines;
using Styx.Helpers;
using Styx.Pathing;
using Styx.Pathing.Avoidance;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using System.Diagnostics;
using Styx.WoWInternals.World;
using Styx.CommonBot.POI;

// ReSharper disable CheckNamespace

namespace Bots.DungeonBuddy.DungeonScripts.Legion
// ReSharper restore CheckNamespace
{
    public class ChamberOfTheAvatar : TombOfSargeras
    {
        #region Overrides of Dungeon

        public override uint DungeonId => 1496;

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            units.RemoveAll(
                ret =>
                {
                    var unit = ret as WoWUnit;
                    if (unit == null)
                        return false;

                    // wait a bit for tank to get threat on boss.
                    if (unit.Entry == MobId_FallenAvatar && unit.Z < 1515 && !_fallenAvatarPhase2TransitionTimer.IsFinished)
                        return true;

                    return false;
                });
        }


        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {

            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit == null)
                    continue;
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            var isDps = Me.IsDps;

            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit == null)
                    continue;

                switch (unit.Entry)
                {
                }
            }
        }

        public override void OnEnter()
        {
            base.OnEnter();
            _capabilityHandler = CapabilityManager.Instance.CreateNewHandle();
            GameWorld.UnitSpellLineOfSightTest += GameWorld_UnitSpellLineOfSightTest;
            TreeHooks.Instance.InsertHook("Dungeonbot_Main", 0, _releaseAfterDeath);
        }

        public override void OnExit()
        {
            base.OnExit();
            if (_capabilityHandler != null)
            {
                CapabilityManager.Instance.Clear(_capabilityHandler);
                _capabilityHandler = null;
            }
            GameWorld.UnitSpellLineOfSightTest -= GameWorld_UnitSpellLineOfSightTest;

            TreeHooks.Instance.RemoveHook("Dungeonbot_Main", _releaseAfterDeath);
        }

        private void GameWorld_UnitSpellLineOfSightTest(object sender, UnitSpellLineOfSightTestEventArgs e)
        {
            if (e.Unit.Entry == MobId_FallenAvatar && e.Unit.Z > 1660 && e.Unit.DistanceSqr < 50 * 50)
                e.Handled = e.InSpellLineOfSight = true;
        }

        #endregion

        #region Navigation

        private ElevatorInfo _elevatorInfo = new ElevatorInfo(
                                                elevatorRadius: 20,
                                                gameObjectId: 268574,
                                                topRestingZ: 2886.861f,
                                                bottomRestingZ: 1650.51f,
                                                topGetOnLoc: new Vector3(6348.021f, -795.5745f, 2887.841f),
                                                topGetOffLoc: new Vector3(6319.82f, -796.6534f, 2887.841f),
                                                bottomGetOnLoc: new Vector3(6348.021f, -795.5745f, 1650.51f),
                                                bottomGetOffLoc: new Vector3(6375.346f, -795.4294f, 1650.126f),
                                                name: "Main Elevator");

        private Vector3 _lastMovetoDestination;

        public override async Task<bool> HandleMovement(MoveToParameters moveToParameters)
        {
            _lastMovetoDestination = moveToParameters.Location;
            if (await BoardElevator(moveToParameters.Location) || await ExitElevator())
                return true;
            return false;
        }

        private async Task<bool> BoardElevator(Vector3 destination)
        {
            WoWGameObject transport = Me.Transport as WoWGameObject;
            if (transport != null && transport.Entry == _elevatorInfo.GameObjectId)
                return false;

            transport = ObjectManager.GetObjectsOfType<WoWGameObject>()
                .FirstOrDefault(g => g.Entry == _elevatorInfo.GameObjectId);
            if (transport == null)
                return false;

            var meOnTopFloor = _elevatorInfo.OnTopFloor(Me.Location);
            var destOnTopFloor = _elevatorInfo.OnTopFloor(destination);

            // My location and destination are on same floor level and we're not on elevator.
            if (meOnTopFloor == destOnTopFloor && !_elevatorInfo.IsPointInElevatorShaft(destination))
                return false;

            bool inRestingPosition = _elevatorInfo.InDesiredRestingPosition(Me.Location, transport.Location);
            var elevatorGetOffLoc = _elevatorInfo.GetExitLocation(Me.Location);
            float roomRadius = meOnTopFloor ? 50 : 33;
            var elevatorGetOnLoc = _elevatorInfo.GetBoardLocation(Me.Location);
            var inRoom = Me.Location.Distance(elevatorGetOnLoc) <= roomRadius;

            Vector3 moveTo = !inRoom
                ? elevatorGetOffLoc
                : (inRestingPosition
                    ? WoWMathHelper.GetRandomPointInCircle(elevatorGetOnLoc, 10)
                    : WoWMathHelper.CalculatePointFrom(Me.Location, elevatorGetOnLoc, 26));

            if (!inRoom || !inRestingPosition)
            {
                if (Navigator.AtLocation(moveTo))
                {
                    TreeRoot.StatusText = "Waiting for elevator to move into position";
                    return true;
                }
                return (await CommonCoroutines.MoveTo(moveTo)).IsSuccessful();
            }

            Navigator.PlayerMover.MoveTowards(moveTo);
            await Coroutine.Wait(4000, () => !Me.IsMoving);
            return true;
        }


        private async Task<bool> ExitElevator()
        {
            WoWGameObject transport = Me.Transport as WoWGameObject;
            if (transport == null || transport.Entry != _elevatorInfo.GameObjectId)
                return false;

            var onDestinationFloor = _elevatorInfo.OnTopFloor(Me.Location) == _elevatorInfo.OnTopFloor(_lastMovetoDestination);

            if (_elevatorInfo.IsPointInElevatorShaft(_lastMovetoDestination) 
                || !_elevatorInfo.InDesiredRestingPosition(_lastMovetoDestination, transport.Location))
            {
                TreeRoot.StatusText = "Waiting on elevator";
                return true;
            }

            Logger.Write("Exiting elevator");
            Navigator.PlayerMover.MoveTowards(_elevatorInfo.GetExitLocation(Me.Location));
            await Coroutine.Wait(4000, () => Me.Transport != transport);
            return true;
        }

        private class ElevatorInfo
        {
            private float _floorLevelDivider;
            public ElevatorInfo(float elevatorRadius, uint gameObjectId, float topRestingZ, float bottomRestingZ, Vector3 topGetOnLoc, Vector3 topGetOffLoc,
                Vector3 bottomGetOnLoc, Vector3 bottomGetOffLoc, string name)
            {
                GameObjectId = gameObjectId;
                TopRestingZ = topRestingZ;
                BottomRestingZ = bottomRestingZ;
                TopGetOnLoc = topGetOnLoc;
                TopGetOffLoc = topGetOffLoc;
                BottomGetOnLoc = bottomGetOnLoc;
                BottomGetOffLoc = bottomGetOffLoc;
                Name = name;
                ElevatorRadius = elevatorRadius;
                _floorLevelDivider = (topRestingZ + bottomRestingZ) / 2;
            }

            public uint GameObjectId { get; }

            public float TopRestingZ { get; }

            public float BottomRestingZ { get; }

            public Vector3 TopGetOffLoc { get; }

            public Vector3 TopGetOnLoc { get; }

            public Vector3 BottomGetOffLoc { get; }

            public Vector3 BottomGetOnLoc { get; }

            public float ElevatorRadius { get; }

            public string Name { get; }

            public bool IsPointInElevatorShaft(Vector3 point)
            {
                return point.Z < TopRestingZ + 5 && point.Z > BottomRestingZ - 5 &&
                    point.Distance2DSquared(TopGetOnLoc) < ElevatorRadius * ElevatorRadius;
            }

            // Make sure we're completely on elevator, not just standing on the edge; which might cause toon to fall off and die.
            public bool IsFullyOnElevator(Vector3 myLoc)
            {
                var boardLoc = GetBoardLocation(Me.Location);
                return IsPointInElevatorShaft(Me.Location) &&
                       Me.Location.Distance2D(boardLoc) <= ElevatorRadius / 2f;
            }

            public Vector3 GetRandomBoardLocation(Vector3 myLoc)
            {
                var boardLoc = GetBoardLocation(myLoc);
                return WoWMathHelper.GetRandomPointInCircle(boardLoc, ElevatorRadius / 2f);
            }

            public Vector3 GetBoardLocation(Vector3 loc) => OnTopFloor(loc) ? TopGetOnLoc : BottomGetOnLoc;

            public Vector3 GetExitLocation(Vector3 loc) => OnTopFloor(loc) ? TopGetOffLoc : BottomGetOffLoc;

            public float GetRestingZ(Vector3 loc) => OnTopFloor(loc) ? TopRestingZ : BottomRestingZ;

            public bool InDesiredRestingPosition(Vector3 myLoc, Vector3 elevatorLoc) => Math.Abs(elevatorLoc.Z - GetRestingZ(myLoc)) < 1;

            public bool OnTopFloor(Vector3 loc) => loc.Z > _floorLevelDivider;
        }

        #endregion

        #region Root

        private ActionRunCoroutine _releaseAfterDeath = new ActionRunCoroutine(ctx => ReleaseAfterDeathHandler());

        private async static Task<bool> ReleaseAfterDeathHandler()
        {
            if (!Me.IsDead || Lua.GetReturnVal<bool>("return IsEncounterInProgress()", 0))
                return false;

            Lua.DoString("RepopMe()");
            return true;
        }

        private CapabilityManagerHandle _capabilityHandler;

        [EncounterHandler(0, "Root Handler")]
        public Func<WoWUnit, Task<bool>> RootBehavior()
        {
            return async npc =>
            {
                if (await ExitElevator())
                    return true;



                return false;
            };
        }

        #endregion

        #region Maiden of Vigilance

        private const uint MobId_FragmentofObliteration = 118643;
        private const uint MobId_FragmentofCreation = 118640;

        private const uint MobId_MaidenOfVigilance = 118289;
        private const int SpellId_WrathoftheCreators = 234891;

        [EncounterHandler(MobId_MaidenOfVigilance, "Maiden of Vigilance")]
        public Func<WoWUnit, Task<bool>> MaidenOfVigilanceBehavior()
        {
            Vector3 elevatorShaftTopCenter = new Vector3(6348.021f, -795.5745f, 2888.268f);

            AddAvoidObject<WoWUnit>(() => Me.HasAura("Light Infusion"), 5, u => u.Entry == MobId_FragmentofObliteration);
            AddAvoidObject<WoWUnit>(() => Me.HasAura("Fel Infusion"), 5, u => u.Entry == MobId_FragmentofCreation);

            return async boss =>
            {
                CapabilityManager.Instance.Update(_capabilityHandler, CapabilityFlags.MoveBehind,
                    () => ScriptHelpers.IsViable(boss) && boss.Combat && boss.Z > 1660);

                var unstableSoulAura = Me.GetAuraByName("Unstable Soul");
                if (unstableSoulAura != null && !Me.IsFalling)
                {
                    if (unstableSoulAura.TimeLeft > TimeSpan.FromSeconds(2))
                    {
                        var moveTo = WoWMathHelper.CalculatePointFrom(Me.Location, elevatorShaftTopCenter, 26);
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => unstableSoulAura != null && unstableSoulAura.TimeLeft > TimeSpan.FromSeconds(2),
                            moveTo, "Unstable Soul Destination", 1);
                    }
                    Navigator.PlayerMover.MoveTowards(elevatorShaftTopCenter);
                    await Coroutine.Wait(2000, () => Me.IsFalling);
                    return true;
                }

                //uint infusionId = Me.HasAura("Light Infusion")
                //    ? MobId_FragmentofCreation
                //    : (Me.HasAura("Fel Infusion") ? MobId_FragmentofObliteration : 0u);

                //if (infusionId != 0)
                //{
                //    var fragment = ObjectManager.GetObjectsOfType<WoWUnit>()
                //        .Where(o => o.Entry == infusionId && o.IsFacing(Me))
                //        .MinByOrDefault(o => o.DistanceSqr);

                //    if (fragment != null)
                //        return (await CommonCoroutines.MoveTo(fragment.Location, fragment.SafeName)).IsSuccessful();

                //    return false;
                //}

                CapabilityManager.Instance.Update(_capabilityHandler, CapabilityFlags.Taunting,
                    () => ScriptHelpers.IsViable(boss) && boss.Combat);

                if (!boss.HasAura("Titanic Bulwark") && await ScriptHelpers.InterruptCast(boss, SpellId_WrathoftheCreators))
                    return true;

                return false;
            };
        }

        #endregion

        #region Fallen Avatar

        private Vector3 _fallenAvatarPhase2EncounterCenter = new Vector3(6605.432f, -792.6328f, 1509.918f);

        private Vector2[][] _lavaSections = new[]
        {
            // South
            new []{ new Vector2(-58.78906f, 61.42902f), new Vector2(-44.87793f, -56.94257f), new Vector2(-99.63428f, -71.11108f), new Vector2(-96.01904f, 96.05524f)},
            // South-east
            new []{ new Vector2(9.440918f, -58.86218f), new Vector2(-53.24121f, -37.88287f), new Vector2(-101.4082f, -66.6875f), new Vector2(-65.89111f, -113.7011f), new Vector2(37.10596f, -109.0207f), },
            // North-east
            new []{ new Vector2(-9.92627f, -56.94446f), new Vector2(47.44971f, -37.27222f), new Vector2(90.68994f, -61.89026f), new Vector2(36.06689f, -108.3845f)},
            // North-north-east
            new []{ new Vector2(28.98486f, -48.612f), new Vector2(53.3999f, 13.86719f), new Vector2(146.7959f, 61.2002f), new Vector2(179.0269f, -25.88727f), new Vector2(91.43994f, -64.2196f), },
            // North-north-west
            new []{new Vector2(52.25f, -6.411072f), new Vector2(24.77002f, 52.2724f), new Vector2(63.17871f, 84.89624f), new Vector2(94.36279f, 24.26813f), },
            // North-west
            new []{ new Vector2(-23.33496f, 51.22601f), new Vector2(36.90186f, 39.89319f), new Vector2(66.06396f, 80.90961f), new Vector2(16.76709f, 100.3755f), new Vector2(-29.69189f, 98.45068f)},
            // South-west
            new []{ new Vector2(-4.273926f, 52.80682f), new Vector2(-66.37012f, 38.66614f), new Vector2(-86.75293f, 37.2251f), new Vector2(-99.52734f, 90.11969f), new Vector2(-26.79834f, 103.0318f)},
        };

        private struct FloorPiece
        {
            public FloorPiece (Vector3 location, Vector2[] poly)
            {
                Location = location;
                Poly = poly;
            }

            public Vector3 Location { get; set; }
            public Vector2[] Poly { get; set; }
        }


        private const int SpellId_RuptureRealities_Phase1 = 239132;
        private const int SpellId_RuptureRealities_Phase2 = 235572;
        private const uint MobId_FallenAvatar = 116939;
        private const uint MobId_CorruptedBlade = 119158;
        private const uint MissileSpellId_UnboundChaos = 239482;
        private const uint GameObjectId_7_2_TombRaidAvatarUpperFloorMRN=267934;
        private WaitTimer _fallenAvatarPhase2TransitionTimer = new WaitTimer(TimeSpan.FromSeconds(10));

        [EncounterHandler(MobId_FallenAvatar, "Fallen Avatar")]
        public Func<WoWUnit, Task<bool>> FallenAvatarBehavior()
        {
            var destructableFloorLocation = new Vector3(6599.8f, -794.1996f, 1508.5f);

            Dictionary<uint, FloorPiece> floorPieces = new Dictionary<uint, FloorPiece>() {
            // Floor #1
            {267709, new FloorPiece(location: new Vector3(6558.186f, -756.8498f, 1509.918f),
                                    poly: new []{new Vector2(9.943848f, 11.367f), new Vector2(-11.2959f, 12.31372f), new Vector2(-14.47217f, -14.10602f), new Vector2(8.925781f, -9.838501f) }) }, 
            // Floor #2
            {267712, new FloorPiece(location: new Vector3(6577.491f, -754.2687f, 1509.918f),
                                    poly: new []{new Vector2(-10.69922f, -13.04303f), new Vector2(-10.22119f, 13.13477f), new Vector2(8.996582f, 14.15088f), new Vector2(10.1626f, -11.49573f)}) }, 
            // Floor #3
            {267711, new FloorPiece(location: new Vector3(6596.573f, -753.6849f, 1509.956f),
                                    poly: new []{new Vector2(11.45557f, 6.829956f), new Vector2(2.227539f, 17.32111f), new Vector2(-13.82324f, 12.33575f), new Vector2(-10.14941f, -12.3786f), new Vector2(12.54053f, -13.54974f)}) }, 
            // Floor #4
            {267714, new FloorPiece(location: new Vector3(6617.424f, -757.6608f, 1509.893f),
                                    poly: new []{new Vector2(8.38916f, -10.21429f), new Vector2(-10.37402f, -8.408691f), new Vector2(-13.34375f, 12.59882f), new Vector2(11.37695f, 9.372498f)}) }, 
            // Floor #5
            {267703, new FloorPiece(location: new Vector3(6635.225f, -761.1328f, 1509.907f),
                                    poly: new []{new Vector2(-6.52002f, 13.54089f), new Vector2(0.2490234f, 13.6424f), new Vector2(17.7168f, 0.4711914f), new Vector2(11.32666f, -9.818665f), new Vector2(-10.63086f, -8.780762f)}) }, 
            // Floor #6
            {267701, new FloorPiece(location: new Vector3(6559.799f, -776.82f, 1509.862f),
                                    poly: new []{new Vector2(-9.883789f, -11.74701f), new Vector2(-12.2749f, 8.246521f), new Vector2(10.96338f, 14.0976f), new Vector2(10.64404f, -8.544922f)}) }, 
            // Floor #7
            {267704, new FloorPiece(location: new Vector3(6578.726f, -774.8391f, 1509.918f) ,
                                    poly: new []{new Vector2(12.28613f, -11.95056f), new Vector2(11.56592f, 13.03088f), new Vector2(-10.09619f, 12.01129f), new Vector2(-10.58301f, -12.22186f)}) }, 
            // Floor #8
            {267702, new FloorPiece(location: new Vector3(6599.142f, -774.3633f, 1509.864f),
                                    poly: new []{new Vector2(-10.39502f, 10.47266f), new Vector2(-12.08594f, -11.15143f), new Vector2(10.396f, -12.02533f), new Vector2(11.45313f, 9.825195f) }) }, 
            // Floor #9
            {267713, new FloorPiece(location: new Vector3(6619.104f, -775.3428f, 1509.905f),
                                    poly: new []{new Vector2(-13.80811f, 11.33765f), new Vector2(-12.97412f, -10.90601f), new Vector2(14.91406f, -12.70654f), new Vector2(11.75f, 10.7182f)}) }, 
            // Floor #10
            {267706, new FloorPiece(location: new Vector3(6641.892f, -777.587f, 1509.893f),
                                    poly: new []{new Vector2(-14.48389f, 9.426392f), new Vector2(-12.14893f, -11.14551f), new Vector2(15.43896f, -11.48944f), new Vector2(8.182129f, 10.34375f)}) }, 
            // Floor #11
            {267707, new FloorPiece(location: new Vector3(6559.712f, -795.7538f, 1509.918f),
                                    poly: new []{new Vector2(11.65527f, 12.4657f), new Vector2(-12.72998f, 11.22998f), new Vector2(-12.09082f, -10.5481f), new Vector2(11.44189f, -11.48834f) }) }, 
            // Floor #12
            {267708, new FloorPiece(location: new Vector3(6580.161f, -794.8896f, 1509.918f),
                                    poly: new []{new Vector2(9.336914f, 13.28162f), new Vector2(-10.68018f, 10.27148f), new Vector2(-9.725098f, -11.94952f), new Vector2(12.04297f, -8.798401f) }) }, 
            // Floor #13
            {267700, new FloorPiece(location: new Vector3(6599.676f, -794.3187f, 1509.918f),
                                    poly: new []{new Vector2(-10.8916f, -10.81219f), new Vector2(-11.68066f, 12.2594f), new Vector2(12.82715f, 11.4881f), new Vector2(10.08398f, -11.62646f)}) }, 
            // Floor #14
            {267710, new FloorPiece(location: new Vector3(6620.699f, -795.9968f, 1509.918f),
                                    poly: new []{new Vector2(10.03271f, -11.44415f), new Vector2(-13.00635f, -10.62195f), new Vector2(-11.56543f, 12.07495f), new Vector2(12.84277f, 10.91785f)}) }, 
            // Floor #15
            {267705, new FloorPiece(location: new Vector3(6641.438f, -796.5995f, 1509.887f),
                                    poly: new []{new Vector2(21.05713f, 7.091003f), new Vector2(-11.54883f, 12.09607f), new Vector2(-12.88086f, -10.6684f), new Vector2(8.757813f, -11.45764f) }) }, 
            // Floor #16
            {267724, new FloorPiece(location: new Vector3(6563.06f, -813.7796f, 1509.813f),
                                    poly: new []{new Vector2(10.521f, -10.84229f), new Vector2(8.277832f, 10.21753f), new Vector2(-10.33203f, 9.572693f), new Vector2(-11.50586f, -9.672485f)}) }, 
            // Floor #17
            {267716, new FloorPiece(location: new Vector3(6580.543f, -813.1677f, 1509.919f),
                                    poly: new []{new Vector2(11.27197f, 11.159f), new Vector2(-12.354f, 12.74292f), new Vector2(-9.098145f, -12.73138f), new Vector2(11.78613f, -12.12769f) }) }, 
            // Floor #18
            {267721, new FloorPiece(location: new Vector3(6599.873f, -814.2256f, 1509.953f),
                                    poly: new []{new Vector2(-10.38086f, 11.81818f), new Vector2(-11.01611f, -11.42212f), new Vector2(8.419922f, -12.90649f), new Vector2(11.71387f, 10.52356f)}) }, 
            // Floor #19
            {267718, new FloorPiece(location: new Vector3(6617.964f, -814.9599f, 1509.941f),
                                    poly: new []{new Vector2(14.06592f, 11.05792f), new Vector2(-10.79785f, 10.95099f), new Vector2(-10.15967f, -12.84198f), new Vector2(9.322266f, -10.10632f)}) }, 
            // Floor #20
            {267719, new FloorPiece(location: new Vector3(6637.677f, -814.2394f, 1509.909f),
                                    poly: new []{new Vector2(9.797363f, -10.8847f), new Vector2(11.88623f, 9.300293f), new Vector2(-9.331543f, 11.71606f), new Vector2(-10.85059f, -10.81244f) }) }, 
            // Floor #21
            {267722, new FloorPiece(location: new Vector3(6565.716f, -829.9774f, 1509.85f),
                                    poly: new []{new Vector2(10.32031f, -12.15601f), new Vector2(10.32031f, 7.564209f), new Vector2(-13.875f, 8.547546f), new Vector2(-6.655762f, -7.750366f)}) }, 
            // Floor #22
            {267723, new FloorPiece(location: new Vector3(6582.522f, -832.8466f, 1509.823f),
                                    poly: new []{new Vector2(-9.779785f, 10.0863f), new Vector2(10.34814f, 10.1723f), new Vector2(8.140137f, -13.9469f), new Vector2(-8.47998f, -8.713501f)}) }, 
            // Floor #23
            {267715, new FloorPiece(location:  new Vector3(6599.16f, -836.7735f, 1509.845f),
                                    poly: new []{new Vector2(-9.718262f, 15.20032f), new Vector2(13.08984f, 15.22693f), new Vector2(7.203613f, -17.81488f), new Vector2(-11.17334f, -13.27618f)}) }, 
            // Floor #24
            {267717, new FloorPiece(location:  new Vector3(6613.758f, -835.6019f, 1509.86f),
                                    poly: new []{new Vector2(-8.678711f, -12.68445f), new Vector2(-5.193848f, 12.73364f), new Vector2(11.76123f, 12.93506f), new Vector2(1.013184f, -13.88226f)}) }, 
            // Floor #25
            {267720, new FloorPiece(location: new Vector3(6629.96f, -832.574f, 1509.865f),
                                    poly: new []{new Vector2(-16.15283f, -16.35272f), new Vector2(-5.011719f, 14.27325f), new Vector2(16.16602f, 7.389099f), new Vector2(5.384277f, -9.890747f)}) },
            };

            WoWUnit fallenAvatar = null;

            Func<WoWPlayer, bool> shouldStackOnDarkMark = player =>
            {
                var aura = player.GetAuraByName("Dark Mark");
                if (aura == null)
                    return false;

                if (aura.TimeLeft >= TimeSpan.FromMilliseconds(2000))
                    return true;

                return Me.RaidMembers.Count(g => !g.IsMe && g != player && g.Location.DistanceSquared(player.Location) < 5 * 5) >= 2;
            };

            Func<WoWGameObject, bool> isDestroyed = gObj => (gObj.FlagsUint & 0x400) != 0;

            PerFrameCachedValue<HashSet<uint>> activeFloorPieces = new PerFrameCachedValue<HashSet<uint>>(
                () => new HashSet<uint>(ObjectManager.GetObjectsOfType<WoWGameObject>()
                    .Where(g => g.SubType == WoWGameObjectType.DestructibleBuilding 
                        && !isDestroyed(g) && g.Location.DistanceSquared(destructableFloorLocation) < 1*1)
                    .Select(g=>g.Entry)));

            var unboundChaosTimer = new WaitTimer(TimeSpan.FromMilliseconds(500));
            WoWPlayer lastUnboundChaosTarget = null;

            // Avoids the regular lava area
            AddAvoidPolygon(() => ScriptHelpers.IsViable(fallenAvatar) && fallenAvatar.Combat,
                null, 40, e => 0, e => 1, e => 40, e => e, e => _fallenAvatarPhase2EncounterCenter,
                () => _lavaSections);

            AddAvoidPolygon(() => ScriptHelpers.IsViable(fallenAvatar) && fallenAvatar.Combat,
                  null, 40, e => 0, e => 1, e => 40, e => e.Value.Poly, e => e.Value.Location,
                  () => floorPieces.Where(f => !activeFloorPieces.Value.Contains(f.Key)));

            AddAvoidLocation(() => true, 5, m => m.ImpactPosition, 
                () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == MissileSpellId_UnboundChaos));

            AddAvoidObject<WoWUnit>(() => true, 60, 
                u => u.Entry == MobId_FallenAvatar 
                    && (u.CastingSpellId == SpellId_RuptureRealities_Phase1 || u.CastingSpellId == SpellId_RuptureRealities_Phase2));

            AddAvoidLine<WoWUnit>(() => true, u => u.Location, u => u.RenderFacing, u => 50, u => 8, u =>
            {
                if (u.Entry != MobId_CorruptedBlade)
                    return false;
                var aura = Me.GetAuraByName("Shadowy Blades");
                return aura == null || u.Guid != aura.CreatorGuid;
            });

            AddAvoidObject<WoWPlayer>(() => true, 8, u => !u.IsMe && (u.HasAura("Shadowy Blades") || Me.HasAura("Shadowy Blades")));
            AddAvoidObject<WoWPlayer>(() => true, 8, u => !u.IsMe && u.HasAura("Dark Mark") && !shouldStackOnDarkMark(u));

            return async boss =>
            {
                fallenAvatar = boss;
                CapabilityManager.Instance.Update(_capabilityHandler, CapabilityFlags.MoveBehind,
                    () => ScriptHelpers.IsViable(boss) && boss.Combat && boss.Z > 1660);

                CapabilityManager.Instance.Update(_capabilityHandler, CapabilityFlags.GapCloser,
                    () => ScriptHelpers.IsViable(boss) && boss.Combat && boss.Z > 1660);

                CapabilityManager.Instance.Update(_capabilityHandler, CapabilityFlags.Taunting,
                    () => ScriptHelpers.IsViable(boss) && boss.Combat && boss.Z > 1660);

                if (boss.Z > 1515)
                    _fallenAvatarPhase2TransitionTimer.Reset();

                var darkMarkTarget = Me.RaidMembers.FirstOrDefault(r => !r.IsMe && shouldStackOnDarkMark(r));

                if (darkMarkTarget != null)
                {
                    return await ScriptHelpers.StayAtLocationWhile(
                        () => ScriptHelpers.IsViable(darkMarkTarget) && darkMarkTarget.HasAura("Dark Mark"),
                        darkMarkTarget.Location, "Dark Mark", 7);
                }


                return false;
            };
        }

        #endregion

    }
}