﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Threading.Tasks;
using Bots.DungeonBuddy.Attributes;
using Bots.DungeonBuddy.Enums;
using Bots.DungeonBuddy.Helpers;
using Bots.DungeonBuddy.Profiles;
using Buddy.Coroutines;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Bars;
using Styx.CommonBot.Coroutines;
using Styx.CommonBot.Routines;
using Styx.Helpers;
using Styx.Pathing;
using Styx.Pathing.Avoidance;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

// ReSharper disable CheckNamespace

namespace Bots.DungeonBuddy.DungeonScripts.Legion
// ReSharper restore CheckNamespace
{
    public abstract class TheEmeraldNightmare: Dungeon
    {
        #region Overrides of Dungeon
        public override Vector3 Entrance { get; } = new Vector3(3615.687f, 6508.461f, 183.4678f);

        public override Vector3 ExitLocation { get; } = new Vector3(1790.058f, 1420.656f, 347.8831f);


        public override void OnEnter()
        {
            if (Me.IsTank())
            {
                Alert.Show(
                    "Tanking Not Supported",
                    string.Format(
                        "Tanking is not supported in the {0} script. " +
                        "If you wish to stay in raid and play manually then press 'Continue', otherwise you will automatically leave raid.",
                        Name),
                    30,
                    true,
                    true,
                    null,
                    () => Lua.DoString("LeaveParty()"),
                    "Continue",
                    "Leave");
            }
            else
            {
                Alert.Show(
                    "Do Not AFK",
                    "It is highly recommended you do not afk while in a raid and be prepared to intervene " +
                    "if needed in the event something goes wrong or you're asked to perform a certain task.",
                    20,
                    true,
                    false,
                    null,
                    null,
                    "Ok");
            }
        }
        #endregion

        protected abstract class DarkboughArea
        {
            public abstract bool IsInArea(Vector3 point);
            public abstract Vector3 PortInLocation { get; }
            public abstract Vector3 PortOutLocation { get; }

            public virtual async Task<bool> HandlePortIn()
            {
                return (await CommonCoroutines.MoveTo(PortInLocation)).IsSuccessful();
            }
        }

        protected class MainArea : DarkboughArea
        {
            public override Vector3 PortInLocation => throw new NotImplementedException();

            public override Vector3 PortOutLocation => throw new NotImplementedException();

            public override bool IsInArea(Vector3 point) => point.X > 1600 && point.X < 2300 && point.Y > 1000 && point.Y < 1800;
        }

        protected static LocalPlayer Me => StyxWoW.Me;
    }

    public class Darkbough : TheEmeraldNightmare
    {
        #region Overrides of Dungeon

        public override uint DungeonId => 1287;

        public override void OnEnter()
        {
            _capabilityHandle = CapabilityManager.Instance.CreateNewHandle();
            base.OnEnter();
        }

        public override void OnExit()
        {
            if (_capabilityHandle != null)
            {
                CapabilityManager.Instance.Clear(_capabilityHandle);
                _capabilityHandle = null;
            }
            base.OnExit();
        }

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            var tanks = Me.RaidMembers.Where(t => t.IsTank).ToList();
            units.RemoveAll(
                ret =>
                {
                    var unit = ret as WoWUnit;
                    if (unit == null)
                        return false;

                    // ignore any trash pulled by stranglers 
                    if (unit.Combat && !tanks.Any(t => t.Location.DistanceSquared(unit.Location) < 100 * 100))
                        return true;

                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            var isTank = Me.IsTank;
            var inDungeon = Me.MapId == LfgDungeon.MapId;

            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit == null)
                    continue;
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            var isDps = Me.IsDps;

            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit == null)
                    continue;
            }
        }

        public override async Task<bool> HandleMovement(MoveToParameters moveToParameters)
        {
            var myLoc = (WoWMovement.ActiveMover ?? Me).Location;
            var dest = moveToParameters.Location;
            var areaAtMyLoc = _darkboughAreas.FirstOrDefault(a => a.IsInArea(myLoc));
            var areaAtDest = _darkboughAreas.FirstOrDefault(a => a.IsInArea(dest));

            if (areaAtMyLoc != null && areaAtDest != null && areaAtMyLoc != areaAtDest)
            {
                if (areaAtMyLoc.GetType() == typeof(MainArea))
                    return await areaAtDest.HandlePortIn();

                return (await CommonCoroutines.MoveTo(areaAtMyLoc.PortOutLocation)).IsSuccessful();
            }

            return false;
        }

        private DarkboughArea[] _darkboughAreas = new DarkboughArea[]
        {
            new MainArea(),
            new IlgynothArea(),
            new EleretheRenferalArea()
        };


        private class IlgynothArea : DarkboughArea
        {
            public override Vector3 PortInLocation { get; } = new Vector3(1796.659f, 1328.852f, 601.0168f);
            public override Vector3 PortOutLocation { get; } = new Vector3(-13149.33f, 12071.88f, -254.3859f);
            public override bool IsInArea(Vector3 point) => point.X > -14382  && point.X < -11206 && point.Y > 10668 && point.Y < 13860;
        }

        private class EleretheRenferalArea : DarkboughArea
        {
            public override Vector3 PortInLocation { get; } = new Vector3(1833.781f, 1254.496f, 600.8725f);
            public override Vector3 PortOutLocation { get; } = new Vector3(11325.6f, 11033.48f, 14.84913f);
            public override bool IsInArea(Vector3 point) => point.X > 8537 && point.X < 12794 && point.Y > 10135 && point.Y < 12797;
        }


        #endregion

        #region Root


        [EncounterHandler(0, "Root Handler")]
        public Func<WoWUnit, Task<bool>> RootBehavior()
        {
            CapabilityManagerHandle handle = CapabilityManager.Instance.CreateNewHandle();
            return async npc =>
            {
                var befoulmentTarget = Me.RaidMembers.FirstOrDefault(u => u.HasAura("Befoulment"));
                if (befoulmentTarget != null)
                {
                    if (!befoulmentTarget.IsMe)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => ScriptHelpers.IsViable(befoulmentTarget) && befoulmentTarget.HasAura("Befoulment"),
                            befoulmentTarget.Location,
                            "Befoulment",
                            4);
                    }
                    else
                    {
                        var nearestRaidMember = Me.RaidMembers.Where(r => !r.IsMe).MinByOrDefault(r => r.DistanceSqr);
                        if (nearestRaidMember != null)
                        {
                            return await ScriptHelpers.StayAtLocationWhile(
                                         () => ScriptHelpers.IsViable(nearestRaidMember) && Me.HasAura("Befoulment"),
                                         nearestRaidMember.Location,
                                         "Befoulment",
                                         4);
                        }
                    }
                }

                return false;
            };
        }

        #endregion

        #region Nythendra

        #region Trash

        private const uint MobId_NightmareHorror = 105591;
        private const uint AreaTriggerId_NightmareFog = 12401;

        [EncounterHandler(MobId_NightmareHorror, "Nightmare Horror")]
        public Func<WoWUnit, Task<bool>> NightmareHorrorHandler()
        {
            AddAvoidObject<WoWAreaTrigger>(7, AreaTriggerId_NightmareFog);
            return npc => Task.FromResult(false);
        }


        private const uint MobId_TaintheartBefouler = 111354;


        private const uint MobId_DeathrootAncient = 111528;
        private const int SpellId_DevastatingStomp = 225249;
        private const int SpellId_RottingGasp = 225296;

        [EncounterHandler(MobId_DeathrootAncient, "Deathroot Ancient")]
        public Func<WoWUnit, Task<bool>> DeathrootAncientHandler()
        {
            AddAvoidObject<WoWUnit>(20, u => u.Entry == MobId_DeathrootAncient && u.CastingSpellId == SpellId_DevastatingStomp);
            AddAvoidUnitCone(0, 20, 60, u => u.Entry == MobId_DeathrootAncient && u.CastingSpellId == SpellId_RottingGasp);

            return async boss =>
            {
                return false;
            };
        }


        private const uint AreaTriggerId_VenomousPool = 11547;
        private const uint MobId_ScreechingSpiderling = 112039;

        [EncounterHandler(MobId_ScreechingSpiderling, "Screeching Spiderling")]
        public Func<WoWUnit, Task<bool>> ScreechingSpiderlingHandler()
        {

            return async boss =>
            {
                return false;
            };
        }

        private const uint MobId_Nightmother = 111980;

        [EncounterHandler(MobId_Nightmother, "Nightmother")]
        public Func<WoWUnit, Task<bool>> NightmotherHandler()
        {
            AddAvoidObject<WoWAreaTrigger>(5, AreaTriggerId_VenomousPool);
            // ToDo. Add avoidance for Wretched Ambush 
            return boss => Task.FromResult(false);
        }

        private const uint MobId_Shadowfeather = 111975;

        CapabilityManagerHandle _capabilityHandle = CapabilityManager.Instance.CreateNewHandle();

        [EncounterHandler(MobId_Shadowfeather, "Shadowfeather ")]
        public Func<WoWUnit, Task<bool>> ShadowfeatherHandler()
        {
            const int spellId_GatheringWinds = 222999;

            return async boss =>
            {
                HandleGatheringWinds(boss, spellId_GatheringWinds);
                return false;
            };
        } 


        #endregion

        private const uint MobId_Nythendra = 102672;
        private const uint AreaTriggerId_InfestedGround = 10594;
        private const int SpellId_InfestedBreath = 202977;
        private const int MissileSpellId_DrakesWrath = 225325;

        [EncounterHandler(MobId_Nythendra, "Nythendra")]
        public Func<WoWUnit, Task<bool>> NythendraHandler()
        {
            AddAvoidObject<WoWAreaTrigger>(7, AreaTriggerId_InfestedGround);
            // Burst of Corruption Volatile Rot
            AddAvoidObject<WoWPlayer>(() => true, p => Me.HasAura("Rot") ? 15:  8, p => !p.IsMe && (Me.HasAura("Rot") || p.HasAura("Rot")));
            AddAvoidObject<WoWPlayer>(() => true, 8, p => !p.IsMe && (Me.HasAura("Volatile Rot") || p.HasAura("Volatile Rot")));
            AddAvoidUnitCone(0, 45, 60, u => u.Entry == MobId_Nythendra && u.Combat && u.CastingSpellId == SpellId_InfestedBreath);
            //AddAvoidUnitCone(180, 40, 60, u => u.Entry == MobId_Nythendra && u.Combat);
            AddAvoidLocation(() => true, 6, m => m.ImpactPosition,
                () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == MissileSpellId_DrakesWrath));

            var roomCenterLoc = new Vector3(1887.15f, 1310.866f, 369.5015f);

            return boss => Task.FromResult(false);
        }

        #endregion

        #region Il'gynoth

        private const uint MobId_Ilgynoth = 105393;
        private const uint MobId_NightmareIchor = 105721;
        private const uint MobId_EyeofIlgynoth = 105906;
        private const uint MobId_TaintedBloodpetal = 112118;

        [EncounterHandler(MobId_Ilgynoth, "Il'gynoth")]
        public Func<WoWUnit, Task<bool>> IlgynothHandler()
        {
            AddAvoidObject<WoWPlayer>(() => true, 8, p => !p.IsMe && (Me.HasAura("Spew Corruption") || p.HasAura("Spew Corruption")));
            // player keeps running into them at entrance of Il'gynoth's area
            AddAvoidObject<WoWPlayer>(() => true, u => u.MyAggroRange + 5, u => u.Entry == MobId_TaintedBloodpetal && !u.Combat);

            return async boss =>
            {
                if (await ScriptHelpers.DispelGroup("Touch of Corruption", ScriptHelpers.PartyDispelType.Magic))
                    return true;
                var fixateAura = Me.GetAuraByName("Fixate");
                WoWUnit fixateAuraCreator;
                if (fixateAura != null && (fixateAuraCreator = ObjectManager.GetObjectByGuid<WoWUnit>(fixateAura.CreatorGuid)) != null)
                {
                    var eye = ObjectManager.GetObjectsOfType<WoWUnit>().FirstOrDefault(u => u.Entry == MobId_EyeofIlgynoth);
                    if (eye != null)
                        return await ScriptHelpers.StayAtLocationWhile(() => ScriptHelpers.IsViable(fixateAuraCreator) && Me.HasAura("Fixate"), eye.Location);
                }

                return false;
            };
        }

        private const uint MobId_DeathglareTentacle = 105322;
        private const int SpellId_MindFlay = 208697;
        private const int SpellId_GroundSlam = 208689;

        [EncounterHandler(MobId_DeathglareTentacle, "Deathglare Tentacle")]
        public Func<WoWUnit, Task<bool>> DeathglareTentacleHandler()
        {
            // Ground Slam does damage in a line that starts at caster and is dirrected at its current target
            AddAvoidLine<WoWUnit>(() => true, u => u.Location, 
                u => u.Rotation, u => 60, u => 10,
                u => u.Entry == MobId_DeathglareTentacle && u.CastingSpellId == SpellId_GroundSlam && u.CurrentTargetGuid != Me.Guid);

            // Stand away from other players if targeted by Ground Slam
            AddAvoidObject<WoWUnit>(
                () => Targeting.Instance.TargetList.Any(t => t.Entry == MobId_DeathglareTentacle
                    && t.CastingSpellId == SpellId_GroundSlam && t.CurrentTargetGuid == Me.Guid),
                8, p => !p.IsMe);

            return async boss => await ScriptHelpers.InterruptCast(boss, SpellId_MindFlay);
        }

        #endregion

        #region Elerethe Renferal

        private const uint MobId_EleretheRenferal = 106087;
        private const uint MobId_PulsingEggSac = 112078;
        private const uint MobId_TwistingShadows1 = 111439;
        private const uint MobId_TwistingShadows2 = 107767;
        private const int MissileSpellId_VileAmbush1 = 215219;
        private const int MissileSpellId_VileAmbush2 = 214339;
        private const uint MobId_VileAmbushStalker = 107969;
        private const uint MobId_FogBeastSpawnerStalker = 111535;

        [EncounterHandler(MobId_EleretheRenferal, "Elerethe Renferal")]
        public Func<WoWUnit, Task<bool>> EleretheRenferalHandler()
        {
            WoWUnit eleretheRenferal = null;
            const int spellId_GatheringWinds = 212707;
            Func<WoWUnit, bool> castingGatheringWinds = unit => unit.HasAura(spellId_GatheringWinds) || unit.CastingSpellId == spellId_GatheringWinds;

            AddAvoidObject<WoWUnit>(() => true, u => u.HasAura("Pulsing Egg Cosmetic") ? 10 : 5, MobId_PulsingEggSac);
            AddAvoidObject<WoWUnit>(8, MobId_FogBeastSpawnerStalker);

            AddAvoidObject<WoWUnit>(() => true, 6, u => u.Entry == MobId_TwistingShadows1 || u.Entry == MobId_TwistingShadows2);
            AddAvoidObject<WoWPlayer>(() => true, 
                p => Me.HasAura("Necrotic Venom") ? 15:7, 
                p => !p.IsMe && (p.HasAura("Necrotic Venom") || Me.HasAura("Necrotic Venom")));

            AddAvoidObject<WoWPlayer>(() => true,
                p => Me.HasAura("Twisting Shadows") ? 15 : 7,
                p => !p.IsMe && (p.HasAura("Twisting Shadows") || Me.HasAura("Twisting Shadows")));

            AddAvoidObject<WoWUnit>(() => ScriptHelpers.IsViable(eleretheRenferal) && eleretheRenferal.HasAura("Vile Ambush"),
                22, u => u.Entry == MobId_VileAmbushStalker);

            return async boss =>
            {
                eleretheRenferal = boss;
                if (boss.Combat && boss.HealthPercent > 90 && await ScriptHelpers.CastHeroism())
                    return true;

                HandleGatheringWinds(boss, spellId_GatheringWinds);

                return false;
            };
        }

        private void HandleGatheringWinds(WoWUnit unit, int gatheringWindsSpellId)
        {
            if (!IsCastingGatheringWinds(unit, gatheringWindsSpellId))
                return;

            CapabilityManager.Instance.Update(_capabilityHandle, CapabilityFlags.Movement, 
                () => unit.IsValid && IsCastingGatheringWinds(unit, gatheringWindsSpellId));

            if (!Me.MovementInfo.MovingForward)
                WoWMovement.Move(WoWMovement.MovementDirection.Forward, 4000);

            if (!Me.IsSafelyFacing(unit, 20))
                unit.Face();
        }

        private bool IsCastingGatheringWinds(WoWUnit unit, int gatheringWindsSpellId) => 
            unit.HasAura(gatheringWindsSpellId) || unit.CastingSpellId == gatheringWindsSpellId;

        #endregion

    }

}