﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Threading.Tasks;
using Bots.DungeonBuddy.Attributes;
using Bots.DungeonBuddy.Enums;
using Bots.DungeonBuddy.Helpers;
using Bots.DungeonBuddy.Profiles;
using Buddy.Coroutines;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Bars;
using Styx.CommonBot.Coroutines;
using Styx.CommonBot.Routines;
using Styx.Helpers;
using Styx.Pathing;
using Styx.Pathing.Avoidance;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Styx.WoWInternals.World;

// ReSharper disable CheckNamespace

namespace Bots.DungeonBuddy.DungeonScripts.Legion
// ReSharper restore CheckNamespace
{
    public class Nightspire : TheNighthold
    {
        #region Overrides of Dungeon

        public override uint DungeonId => 1292;

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            var tanks = GroupMember.GroupMembers.Where(g => g.IsTank && g.IsAlive).ToArray();
            units.RemoveAll(
                ret =>
                {
                    var unit = ret as WoWUnit;
                    if (unit == null)
                        return false;

                    // It's common for people to get lost and pull mobs in this dungeon...
                    if (unit.Combat && unit.IsTargetingMyRaidMember && !tanks.Any(t => t.Location.DistanceSquared(unit.Location) < 100 * 100))
                        return true;

                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            var isTank = Me.IsTank;
            var inDungeon = Me.MapId == LfgDungeon.MapId;

            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit == null)
                    continue;
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            var isDps = Me.IsDps;

            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit == null)
                    continue;

                switch (unit.Entry)
                {
                    case MobId_FelswornSpellguard:
                        priority.Score += 4500;
                        break;
                    case MobId_PhantasmalBloodfang:
                    case MobId_RecursiveElemental:
                    case MobId_ExpedientElemental:
                        priority.Score += 5000;
                        break;
                    case MobId_TaintedBlood:
                        priority.Score += 5500;
                        break;
                }

            }
        }

        public override void OnEnter()
        {
            base.OnEnter();
            _capabilityHandle = CapabilityManager.Instance.CreateNewHandle();
        }

        public override void OnExit()
        {
            base.OnExit();
            CapabilityManager.Instance.Clear(_capabilityHandle);
            _capabilityHandle = null;
        }

        public override void IncludeLootTargetsFilter(List<WoWObject> incomingObjects, HashSet<WoWObject> outgoingObjects)
        {
            foreach (var lootTarget in incomingObjects)
            {
                var gObj = lootTarget as WoWGameObject;
                if (gObj != null && gObj.Entry == GameObjectId_FelboundChest && gObj.CanLoot &&
                    DungeonBuddySettings.Instance.LootMode != LootMode.Off)
                {
                    outgoingObjects.Add(gObj);
                }
            }
        }

        private const uint MobId_ImageofKhadgar = 110677;
        public override async Task<bool> HandleMovement(MoveToParameters moveToParameters)
        {
            if (ScriptHelpers.IsViable(_krosus) && _krosus.Combat 
                && moveToParameters.Location.DistanceSquared(_krosus.Location) < 3*3)
            {
                var moveTo = WoWMathHelper.CalculatePointFrom(Me.Location, _krosus.Location, 23);
                return (await CommonCoroutines.MoveTo(moveTo)).IsSuccessful();
            }

            var meAtKrosus = IsAtKrosus(Me.Location);
            var destAtKrosus = IsAtKrosus(moveToParameters.Location);
            // Need to talk to Khadgar to teleport out of area
            if (meAtKrosus && !destAtKrosus)
            {
                var imageofKhadgar = ObjectManager.GetObjectsOfType<WoWUnit>().FirstOrDefault(u => u.Entry == MobId_ImageofKhadgar);
                if (imageofKhadgar != null)
                    return await ScriptHelpers.TalkToNpc(imageofKhadgar);
            }

            var meAtElisande = Me.Z > 200;
            var destAtElisande = moveToParameters.Location.Z > 200;
            if (meAtElisande != destAtElisande)
            {
                var portalLoc = !meAtElisande ? _bottomPortalLocation : _topPortalLocation;
                var portalId = !meAtElisande ? MobId_BottomPortal : MobId_TopPortal;
                var portal = ObjectManager.GetObjectsOfType<WoWUnit>()
                    .FirstOrDefault(u => u.Entry == portalId && u.Location.DistanceSquared(portalLoc) < 2 * 2);

                if (portal == null)
                    return (await CommonCoroutines.MoveTo(portalLoc, "Portal")).IsSuccessful();

                return await ScriptHelpers.TalkToNpc(portal, 1);
            }
            return false;
        }

        #endregion

        #region Root

        private CapabilityManagerHandle _capabilityHandle;

        private const uint AreaTriggerId_InfiniteAbyss = 12638;

        [EncounterHandler(0, "Root Handler")]
        public Func<WoWUnit, Task<bool>> RootBehavior()
        {
            AddAvoidObject<WoWAreaTrigger>(4, AreaTriggerId_InfiniteAbyss);
            return async npc =>
            {
                return false;
            };
        }

        #endregion

        #region Krosus

        Vector2[] _krosusPoly = new[]
        {
            new Vector2(-45.31763f, 2913.359f),
            new Vector2(19.40513f, 2782.177f),
            new Vector2(-114.2775f, 2696.826f),
            new Vector2(-188.7932f, 2837.671f),
        };

        private bool IsAtKrosus(Vector3 point) => WoWMathHelper.IsPointInPoly(point, _krosusPoly);

        private const uint MobId_Krosus = 101002;
        private const uint AreaTriggerId_Slam = 10826;
        private const int SpellId_FelBeam = 205370;
        private const int SpellId_Slam = 205862;
        private const int SpellId_OrbOfDestruction = 205361;

        private const int MissileSpellId_BurningPitch = 206281;
        private const uint GameObjectId_FelboundChest = 248513;

        WoWUnit _krosus = null;

        [EncounterHandler(MobId_Krosus, "Krosus", Mode = CallBehaviorMode.Proximity, BossRange = 120)]
        public Func<WoWUnit, Task<bool>> KrosusHandler()
        {
            var felBeamRight = true;
            var felBeamTimer = WaitTimer.TenSeconds;

            // Commented out until we figure out which hand is casting spell.
            //AddAvoidLine<WoWUnit>(() => true,
            //    u => WoWMathHelper.CalculatePointAtSide(u.Location, u.Rotation, 15, felBeamRight),
            //    u => u.Rotation, u => 120, u => 30, u => u.Entry == MobId_Krosus
            //        && u.CastingSpellId == SpellId_FelBeam);

            AddAvoidObject<WoWAreaTrigger>(4, AreaTriggerId_Slam);

            #region Bridges 

            int destroyedBridges = 0;
            int slamCount = 0;
            WaitTimer slamTimer = WaitTimer.FiveSeconds;

            var bridge1Poly = new[]
            {
                new Vector2(-37.21924f, -46.56201f),
                new Vector2(-58.38879f, -8.781006f),
                new Vector2(-93.94625f, -25.12183f),
                new Vector2(-71.03245f, -71.86792f),
            };

            var bridge2Poly = new[]
            {
                new Vector2(-10.49564f, -31.0249f),
                new Vector2(-32.68251f, 8.161133f),
                new Vector2(-68.48685f, -12.26001f),
                new Vector2(-45.21839f, -53.17383f),
            };

            var bridge3Poly = new[]
            {
                new Vector2(11.3823f, -18.90186f),
                new Vector2(-8.916039f, 20.46802f),
                new Vector2(-45.21394f, 1.388184f),
                new Vector2(-17.88281f, -38.99585f),
            };

            var destructableBridgeLoc = new Vector3(-34.88195f, 2830.159f, 0);
            AddAvoidPolygon(() => destroyedBridges > 0, null, 40, v => 0, v => 1, v => 40, v => bridge1Poly, v => v, () => new[] { destructableBridgeLoc });
            AddAvoidPolygon(() => destroyedBridges > 1, null, 40, v => 0, v => 1, v => 40, v => bridge2Poly, v => v, () => new[] { destructableBridgeLoc });
            AddAvoidPolygon(() => destroyedBridges > 2, null, 40, v => 0, v => 1, v => 40, v => bridge3Poly, v => v, () => new[] { destructableBridgeLoc });

            #endregion

            var safeBridgeLoc = new Vector3(-15.04017f, 2842.633f, 3.161343f);

            return async boss =>
            {
                _krosus = boss;
                if (!boss.Combat)
                {
                    destroyedBridges = slamCount = 0;
                    felBeamRight = true;
                    return false;
                }

                CapabilityManager.Instance.Update(_capabilityHandle, CapabilityFlags.GapCloser, () => ScriptHelpers.IsViable(boss) && boss.Combat);
                CapabilityManager.Instance.Update(_capabilityHandle, CapabilityFlags.MoveBehind, () => ScriptHelpers.IsViable(boss) && boss.Combat);

                if (boss.CastingSpellId == SpellId_Slam && slamTimer.IsFinished)
                {
                    slamCount++;
                    // a bridge gets destroyed on every 3rd slam.
                    if (slamCount == 3)
                    {
                        destroyedBridges++;
                        slamCount = 0;
                        Logger.WriteDebug("Bridge getting destroyed, Run Back!");
                    }
                    slamTimer.Reset();
                }

                if (boss.CastingSpellId == SpellId_FelBeam && felBeamTimer.IsFinished)
                {
                    felBeamRight = !felBeamRight;
                    felBeamTimer.Reset();
                }


                if (boss.CastingSpellId == SpellId_OrbOfDestruction && boss.CurrentTargetGuid == Me.Guid)
                    return (await CommonCoroutines.MoveTo(safeBridgeLoc, "Orb of Destruction")).IsSuccessful();

                // Most LFR don't even bother handling these..
                //if (Me.IsRangedDps)
                //{
                //    var burningPitch = WoWMissile.InFlightMissiles
                //        .Where(m => m.SpellId == MissileSpellId_BurningPitch
                //            && !Me.RaidMembers.Any(r => !r.IsMe && r.Location.DistanceSquared(m.ImpactPosition) < 3 * 3))
                //        .MinByOrDefault(m => m.ImpactPosition.DistanceSquared(Me.Location));

                //    if (burningPitch != null)
                //    {
                //        return await ScriptHelpers.StayAtLocationWhile(
                //            () => burningPitch != null, burningPitch.ImpactPosition, "Burning Pitch", 2);
                //    }
                //}
                return false;
            };
        }

        #endregion

        #region Tichondrius

        private const uint MobId_Tichondrius = 103685;
        private const uint MobId_TaintedBlood = 108934;
        private const uint MobId_FelSpire = 108625;
        private const uint MobId_PhantasmalBloodfang = 104326;
        private const uint MobId_CarrionNightmare = 108739;
        private const uint MobId_FelswornSpellguard = 108591;

        private const int SpellId_SeekerSwarm = 213238;
        private const int SpellId_EchoesoftheVoid = 213531;
        private const int SpellId_CarrionNightmare = 215988;

        private const uint AreaTriggerId_EssenceofNight = 10926;

        [EncounterHandler(MobId_Tichondrius, "Tichondrius")]
        public Func<WoWUnit, Task<bool>> TichondriusHandler()
        {
            WoWUnit tichondrius = null;
            AddAvoidObject<WoWUnit>(3, MobId_CarrionNightmare);
            AddAvoidLine<WoWUnit>(() => true, u => u.Location, u => u.Rotation, u => 200, u => 6, 
                u => u.Entry == MobId_CarrionNightmare && u.CastingSpellId == SpellId_CarrionNightmare);

            AddAvoidObject<WoWUnit>(6, MobId_FelSpire);
            AddAvoidObject<WoWPlayer>(
                () => ScriptHelpers.IsViable(tichondrius) && tichondrius.CastingSpellId == SpellId_SeekerSwarm && Me.HasAura("Carrion Plague"),
                15, p => !p.IsMe && !p.HasAura("Carrion Plague"));

            // Avoid standing between boss and players with Carrion Plague to prevent from spreading to self.
            AddAvoidLine<WoWPlayer>(
                () => ScriptHelpers.IsViable(tichondrius) && tichondrius.CastingSpellId == SpellId_SeekerSwarm && !Me.HasAura("Carrion Plague"),
                p => p.Location, p => WoWMathHelper.CalculateNeededFacing(Me.Location, tichondrius.Location),
                p => p.Location.Distance(tichondrius.Location), p => 8, p => p.HasAura("Carrion Plague"));

            PerFrameCachedValue<bool> castingEchosOfTheVoid = new PerFrameCachedValue<bool>(
                () => ScriptHelpers.IsViable(tichondrius) 
                    && (tichondrius.CastingSpellId == SpellId_EchoesoftheVoid || tichondrius.HasAura(SpellId_EchoesoftheVoid)));

            Func<WoWUnit, WoWUnit, Vector3> getFelSpireLosLocation = (boss, spire) => 
                WoWMathHelper.CalculatePointFrom(boss.Location, spire.Location, -8);

            return async boss =>
            {
                tichondrius = boss;
                if (castingEchosOfTheVoid.Value)
                {
                    var felSpireLosLoc = ObjectManager.GetObjectsOfType<WoWUnit>()
                        .Where(u => u.Entry == MobId_FelSpire)
                        .Select(u => getFelSpireLosLocation(boss, u))
                        .MinByOrDefault(u => u.DistanceSquared(Me.Location));

                    if (felSpireLosLoc != Vector3.Zero)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => castingEchosOfTheVoid.Value, felSpireLosLoc, "Echos of the Void LOS");
                    }
                }

                if (!Me.HasAura("Essence of Night"))
                {
                    var essenceofNight = ObjectManager.GetObjectsOfType<WoWAreaTrigger>()
                        .Where(a => a.Entry == AreaTriggerId_EssenceofNight && a.DistanceSqr < 40*40)
                        .MinByOrDefault(a => a.DistanceSqr);

                    if (essenceofNight != null)
                        return (await CommonCoroutines.MoveTo(essenceofNight.Location, essenceofNight.SafeName)).IsSuccessful();
                }
                // ToDO: Handle Volatile Wound/Nether Zone if it exists in LFR 
                return false;
            };
        }

        #endregion

        #region Elisande

        private const uint MobId_Elisande = 106643;
        private const uint AreaTriggerId_SpanningSingularity = 11204;
        private const uint AreaTriggerId_ArcaneticRing = 11109;
        private const uint AreaTriggerId_TimeElementalDiesSlow = 11124;
        private const int MissileSpellId_EpochericOrb = 210275;

        private Vector3 _topPortalLocation = new Vector3(365.6129f, 3052.33f, 216.5816f);
        private Vector3 _bottomPortalLocation = new Vector3(314.1285f, 3104.191f, 146.8835f);

        private const uint MobId_TopPortal = 116667;
        private const uint MobId_BottomPortal = 116662;
        private const int SpellId_DelphuricBeam = 214278;

        [EncounterHandler(MobId_Elisande, "Elisande")]
        public Func<WoWUnit, Task<bool>> ElisandeHandler()
        {
            AddAvoidObject<WoWAreaTrigger>(10, AreaTriggerId_TimeElementalDiesSlow);
            AddAvoidObject<WoWAreaTrigger>(() => Me.HealthPercent < 90, 8, a => a.Entry == AreaTriggerId_SpanningSingularity);
            AddAvoidObject<WoWUnit>(() => Me.HealthPercent < 90, 8, u => u.Entry == MobId_Elisande);
            PerFrameCachedValue<WoWUnit> delphuricBeamTarget = new PerFrameCachedValue<WoWUnit>(() => Me.RaidMembers.FirstOrDefault(r => r.HasAura("Delphuric Beam")));

            AddAvoidLine<WoWUnit>(() => delphuricBeamTarget.Value != null && delphuricBeamTarget.Value != Me,
                u => u.Location, u => WoWMathHelper.CalculateNeededFacing(u.Location, delphuricBeamTarget.Value.Location),
                u => u.Location.Distance(delphuricBeamTarget.Value.Location), u => 8,
                u => u.Entry == MobId_Elisande && u.CastingSpellId == SpellId_DelphuricBeam);
            
            // These do to little damage on LFR to bother avoiding.
            // AddAvoidObject<WoWAreaTrigger>(() => true, 6, a => a.Entry == AreaTriggerId_ArcaneticRing);

            return async boss =>
            {
                if (Me.HasAura("Delphuric Beam"))
                {
                    return await ScriptHelpers.StayAtLocationWhile(
                        () => Me.HasAura("Delphuric Beam") && ScriptHelpers.IsViable(boss), 
                        boss.Location, "Delphuric Beam wait position");
                }

                var epochericOrb = WoWMissile.InFlightMissiles
                    .Where(m => m.SpellId == MissileSpellId_EpochericOrb && m.ImpactPosition.DistanceSquared(Me.Location) < 30 * 30)
                    .MinByOrDefault(m => m.ImpactPosition.DistanceSquared(Me.Location));

                if (epochericOrb != null)
                {
                    return await ScriptHelpers.StayAtLocationWhile(() => ScriptHelpers.IsViable(boss) && boss.Combat && epochericOrb != null,
                        epochericOrb.ImpactPosition, "Epocheric Orb", 2);
                }

                if (!Me.HasAura("Fast Time"))
                {
                    // ToDo - Get Fast Time
                }

                return false;
            };
        }


        private const uint MobId_RecursiveElemental = 105299;
        private const int SpellId_Blast = 221864;

        [EncounterHandler(MobId_RecursiveElemental, "Recursive Elemental")]
        public Func<WoWUnit, Task<bool>> RecursiveElementalHandler()
        {
            return async npc => await ScriptHelpers.InterruptCast(npc, SpellId_Blast);
        }

        private const uint MobId_ExpedientElemental = 105301;
        private const int SpellId_ExothermicRelease = 209568;

        [EncounterHandler(MobId_ExpedientElemental, "Expedient Elemental")]
        public Func<WoWUnit, Task<bool>> ExpedientElementalHandler()
        {
            return async npc => await ScriptHelpers.InterruptCast(npc, SpellId_ExothermicRelease);
        }
        #endregion

    }
}