﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Threading.Tasks;
using Bots.DungeonBuddy.Attributes;
using Bots.DungeonBuddy.Enums;
using Bots.DungeonBuddy.Helpers;
using Bots.DungeonBuddy.Profiles;
using Buddy.Coroutines;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Bars;
using Styx.CommonBot.Coroutines;
using Styx.CommonBot.Routines;
using Styx.Helpers;
using Styx.Pathing;
using Styx.Pathing.Avoidance;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

// ReSharper disable CheckNamespace

namespace Bots.DungeonBuddy.DungeonScripts.Legion
// ReSharper restore CheckNamespace
{
    public class RiftOfAln : TheEmeraldNightmare
    {
        #region Overrides of Dungeon

        public override uint DungeonId => 1289;

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            units.RemoveAll(
                ret =>
                {
                    var unit = ret as WoWUnit;
                    if (unit == null)
                        return false;

                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            var isTank = Me.IsTank;
            var inDungeon = Me.MapId == LfgDungeon.MapId;

            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit == null)
                    continue;
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            if (Me.IsHealer)
                return;

            var isRanged = Me.IsRanged;
            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit == null)
                    continue;

                switch(unit.Entry)
                {
                    case MobId_InconceivableHorror:
                        if (ScriptHelpers.IsViable(_xavius))
                        {
                            // melee should go after horrors near boss and ranged go after those further away.
                            if (isRanged)
                                priority.Score += (5500 + _xavius.Location.Distance(unit.Location));
                            else
                                priority.Score += (5500 - _xavius.Location.Distance(unit.Location));
                        }
                        break;
                }
            }
        }

        public override void OnEnter()
        {
            base.OnEnter();
            _capabilityHandle = CapabilityManager.Instance.CreateNewHandle();
            // We need to control dispeling to manage corruption
            CapabilityManager.Instance.Update(_capabilityHandle, CapabilityFlags.DefensiveDispel, 
                () => Me.GroupInfo.LfgDungeonId == LfgDungeon.Id);
        }

        public override void OnExit()
        {
            base.OnExit();
            CapabilityManager.Instance.Clear(_capabilityHandle);
            _capabilityHandle = null;
        }

        #endregion

        #region Root

        private CapabilityManagerHandle _capabilityHandle;

        private const uint AreaTriggerId_ToxicChitin = 10742;
        [EncounterHandler(0, "Root Handler")]
        public Func<WoWUnit, Task<bool>> RootBehavior()
        {
            AddAvoidObject<WoWAreaTrigger>(AreaTriggerId_ToxicChitin);

            return async npc =>
            {
                return false;
            };
        }

        #endregion

        #region Xavius

        private const uint MobId_Xavius = 103769;
        private const uint MobId_InconceivableHorror = 105611;
        private const uint MobId_LurkingTerror = 103694;
        private const int MissileSpellId_CorruptionMeteor = 206341;

        private const uint AreaTriggerId_TaintedDischarge = 11086;

        private WoWUnit _xavius;

        [EncounterHandler(MobId_Xavius, "Xavius")]
        public Func<WoWUnit, Task<bool>> XaviusHandler()
        {
            PerFrameCachedValue<bool> inDreamState = new PerFrameCachedValue<bool>(() => Me.HasAura("Dream Simulacrum"));

            AddAvoidObject<WoWUnit>(6, MobId_LurkingTerror);
            AddAvoidObject<WoWAreaTrigger>(() => !Me.HasAura("Dream Simulacrum"), 10, a => a.Entry == AreaTriggerId_TaintedDischarge);
            AddAvoidObject<WoWPlayer>(() => true, 14, p => !p.IsMe && (p.HasAura("Nightmare Blades") || Me.HasAura("Nightmare Blades")));
            AddAvoidObject<WoWPlayer>(() => ScriptHelpers.IsViable(_xavius) && _xavius.HealthPercent <= 30,
                16, p => !p.IsMe  && ((p.HasAura("Corruption Meteor") && !Me.HasAura("Dream Simulacrum"))
                || (Me.HasAura("Corruption Meteor") && !p.HasAura("Dream Simulacrum"))));

            AddAvoidObject<WoWPlayer>(() => true, 25, p => !p.IsMe && (p.HasAura("Darkening Soul")));
            
            return async boss =>
            {
                _xavius = boss;
                if (inDreamState.Value || boss.HealthPercent <= 30)
                {
                    var dispelTarget = Me.RaidMembers
                        .Where(r => (r.HasAura("Blackening Soul") || r.HasAura("Darkening Soul"))
                            && !Me.RaidMembers.Any(g => g != r && g.IsAlive && g.Location.DistanceSquared(r.Location) < 10 * 10))
                        .MinByOrDefault(r => r.DistanceSqr);
                    if (dispelTarget != null)
                        return await ScriptHelpers.DispelFriendlyUnit("", ScriptHelpers.PartyDispelType.Magic, dispelTarget);
                }

                if (inDreamState.Value)
                {
                    var corruptionMeteorTarget = Me.RaidMembers.FirstOrDefault(r => !r.IsMe && r.HasAura("Corruption Meteor"));
                    if (corruptionMeteorTarget != null)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => ScriptHelpers.IsViable(corruptionMeteorTarget) && corruptionMeteorTarget.HasAura("Corruption Meteor"),
                            corruptionMeteorTarget.Location, "Corruption Meteor", 14);
                    }

                    var taintedDischarge = ObjectManager.GetObjectsOfType<WoWAreaTrigger>()
                        .Where(a => a.Entry == AreaTriggerId_TaintedDischarge)
                        .MinByOrDefault(a => a.DistanceSqr);

                    if (taintedDischarge != null)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => ScriptHelpers.IsViable(taintedDischarge),
                            taintedDischarge.Location, taintedDischarge.SafeName, 8);
                    }
                }
                else
                {
                    var tormentingFixationAura = Me.GetAuraByName("Tormenting Fixation");
                    if (tormentingFixationAura != null)
                    {
                        var lurkingTerror = ObjectManager.GetObjectByGuid<WoWUnit>(tormentingFixationAura.CreatorGuid);
                        var nearestDreamStatePlayer = Me.RaidMembers.Where(r => r.HasAura("Dream Simulacrum")).MinByOrDefault(r => r.DistanceSqr);
                        if (lurkingTerror != null && nearestDreamStatePlayer != null)
                        {
                            var moveTo = WoWMathHelper.CalculatePointFrom(lurkingTerror.Location, nearestDreamStatePlayer.Location, -4);
                            return await ScriptHelpers.StayAtLocationWhile(
                                () => Me.HasAura("Tormenting Fixation"),
                                moveTo, "Tormenting Fixation", 1);
                        }
                    }
                }



                return false;
            };
        }

        #endregion
    }

}