﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Threading.Tasks;
using Bots.DungeonBuddy.Attributes;
using Bots.DungeonBuddy.Enums;
using Bots.DungeonBuddy.Helpers;
using Bots.DungeonBuddy.Profiles;
using Buddy.Coroutines;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Bars;
using Styx.CommonBot.Coroutines;
using Styx.CommonBot.Routines;
using Styx.Helpers;
using Styx.Pathing;
using Styx.Pathing.Avoidance;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;

// ReSharper disable CheckNamespace

namespace Bots.DungeonBuddy.DungeonScripts.Legion
// ReSharper restore CheckNamespace
{
    public class RoyalAthenaeum : TheNighthold
    {
        #region Overrides of Dungeon

        public override uint DungeonId => 1291;

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            units.RemoveAll(
                ret =>
                {
                    var unit = ret as WoWUnit;
                    if (unit == null)
                        return false;

                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            var isTank = Me.IsTank;
            var inDungeon = Me.MapId == LfgDungeon.MapId;

            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit == null)
                    continue;
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            var isDps = Me.IsDps;

            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit == null)
                    continue;

                switch (unit.Entry)
                {
                    case MobId_IcyEnchantment:
                    case MobId_ArcaneEnchantment:
                    case MobId_FieryEnchantment:
                    case MobId_ParasiticLasher:
                        priority.Score += 5000;
                        break;

                    case MobId_PlasmaSphere:
                    case MobId_ThingThatShouldNotBe:
                        priority.Score += 5500;
                        break;
                }
            }
        }


        #endregion

        #region Root

        private const uint AreaTriggerId_PoisonBrambles = 12724;

        [EncounterHandler(0, "Root Handler")]
        public Func<WoWUnit, Task<bool>> RootBehavior()
        {
            AddAvoidObject<WoWAreaTrigger>(4.5f, AreaTriggerId_PoisonBrambles);

            return async npc =>
            {
                return false;
            };
        }

        #endregion


        #region Spellblade Aluriel

        private const uint MobId_IcyEnchantment = 107237;
        private const uint MobId_FieryEnchantment = 107285;
        private const uint MobId_ArcaneEnchantment = 107287;
        private const uint MobId_SpellbladeAluriel = 104881;
        private const uint AreaTriggerId_DetonateMarkofFrost=11496;
        private const uint AreaTriggerId_DetonateSearingBrand=11556;
        private const int SpellId_Annihilate = 212492;

        [EncounterHandler(MobId_SpellbladeAluriel, "Spellblade Aluriel")]
        public Func<WoWUnit, Task<bool>> SpellbladeAlurielHandler()
        {
            AddAvoidObject<WoWPlayer>(() => true, 8, p => !p.IsMe && (p.HasAura("Searing Brand") || Me.HasAura("Searing Brand")));
            AddAvoidUnitCone(0, 25, 60, u => u.Entry == MobId_SpellbladeAluriel && u.CastingSpellId == SpellId_Annihilate);

            AddAvoidObject<WoWPlayer>(() => true, 8, p => !p.IsMe && (p.HasAura("Mark of Frost") || Me.HasAura("Mark of Frost")));
            AddAvoidObject<WoWAreaTrigger>(5, AreaTriggerId_DetonateMarkofFrost);
            AddAvoidObject<WoWAreaTrigger>(7, AreaTriggerId_DetonateSearingBrand);

            return async boss =>
            {
                // TODO find Detonate: Arcane Orb  and Arcane Orb IDs

                // It's probably fine to just wait for Mark of Frost to expire.
                //if (Me.HasAura("Mark of Frost", a => a.StackCount >= 15))
                //{
                //    var otherMarkOfFrostTarget = Me.RaidMembers.Where(r => !r.IsMe && r.HasAura("Mark of Frost")).MinByOrDefault(r => r.DistanceSqr);
                //    if (otherMarkOfFrostTarget != null)
                //        return (await CommonCoroutines.MoveTo(otherMarkOfFrostTarget.Location, "Mark of Frost meetup")).IsSuccessful();
                //}
                return false;
            };
        }

        #endregion


        #region High Botanist Tel'arn

        private const uint MobId_ParasiticLasher = 109075;
        private const uint MobId_PlasmaSphere = 109804;
        private const uint MobId_HighBotanistTelarn = 104528;
        private const uint MobId_ToxicSpore = 110125;

        private const int MissileSpellId_SolarCollapse = 218152;
        private const int MissileSpellId_ToxicSpores = 219233;

        private const uint AreaTriggerId_ControlledChaosVisual = 10202;

        [EncounterHandler(MobId_HighBotanistTelarn, "High Botanist Tel'arn")]
        public Func<WoWUnit, Task<bool>> HighBotanistTelarnHandler()
        {
            AddAvoidLocation(() => true, 4, m => m.ImpactPosition, () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == MissileSpellId_SolarCollapse));
            AddAvoidLocation(() => true, 4, m => m.ImpactPosition, () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == MissileSpellId_ToxicSpores));
            AddAvoidObject<WoWAreaTrigger>(10, AreaTriggerId_ControlledChaosVisual);
            AddAvoidObject<WoWUnit>(4, MobId_ToxicSpore);
            AddAvoidObject<WoWPlayer>(() => Me.HasAura("Call of Night"), 5, p => !p.IsMe && p.HasAura("Call of Night"));

            return async boss =>
            {
                if (Me.HasAura("Call of Night"))
                {
                    var nearbyUnaffectedTarget = Me.RaidMembers.Where(r => !r.IsMe && !r.HasAura("Call of Night") &&
                            !Me.RaidMembers.Any(g => !g.IsMe && g.HasAura("Call of Night") && g.Location.DistanceSquared(r.Location) <= 5*5))
                        .MinByOrDefault(r => r.DistanceSqr);

                    if (nearbyUnaffectedTarget != null)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => Me.HasAura("Call of Night") && ScriptHelpers.IsViable(nearbyUnaffectedTarget), 
                            nearbyUnaffectedTarget.Location, "Call of Night meetup", 4);
                    }
                }
                return false;
            };
        }

        #endregion

        #region Star Augur Etraeus

        private const uint MobId_StarAugurEtraeus = 103758;
        private const uint AreaTriggerId_Felflame = 10916;

        [EncounterHandler(MobId_StarAugurEtraeus, "Star Augur Etraeus")]
        public Func<WoWUnit, Task<bool>> StarAugurEtraeusHandler()
        {
            AddAvoidObject<WoWPlayer>(() =>true, 10, 
                p => !p.IsMe && (p.HasAura("Icy Ejection", a => a.TimeLeft < TimeSpan.FromSeconds(2)) || Me.HasAura("Icy Ejection")));

            AddAvoidObject<WoWPlayer>(() => true, 10,
                p => !p.IsMe && (p.HasAura("Fel Ejection") || Me.HasAura("Fel Ejection")));

            // Gravitaional pull ToDO check this
            //AddAvoidObject<WoWPlayer>(() => true, 10, p => !p.IsMe && p.HasAura("Gravitational Pull"));
            AddAvoidObject<WoWAreaTrigger>(4f, AreaTriggerId_Felflame);

            return async boss =>
            {
                return false;
            };
        }

        private const int SpellId_WitnesstheVoid = 207720;
        private const uint MobId_ThingThatShouldNotBe = 104880;
        [EncounterHandler(MobId_ThingThatShouldNotBe, "Thing That Should Not Be")]
        public Func<WoWUnit, Task<bool>> ThingThatShouldNotBeHandler()
        {
            return async boss =>
            {
                // Face away to avoid getting feared
                if (boss.CastingSpellId == SpellId_WitnesstheVoid && boss.CurrentCastTimeLeft < TimeSpan.FromSeconds(2))
                {
                    var faceDir = WoWMathHelper.NormalizeRadian(WoWMathHelper.CalculateNeededFacing(Me.Location, boss.Location) + (float)Math.PI);
                    Me.SetFacing(faceDir);
                    await Coroutine.Wait(2000, () => ScriptHelpers.IsViable(boss) && boss.CastingSpellId != SpellId_WitnesstheVoid);
                }
                return false;
            };
        }
        #endregion    
    }
}