﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Threading.Tasks;
using Bots.DungeonBuddy.Attributes;
using Bots.DungeonBuddy.Enums;
using Bots.DungeonBuddy.Helpers;
using Bots.DungeonBuddy.Profiles;
using Buddy.Coroutines;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Bars;
using Styx.CommonBot.Coroutines;
using Styx.CommonBot.Routines;
using Styx.Helpers;
using Styx.Pathing;
using Styx.Pathing.Avoidance;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using System.Diagnostics;

// ReSharper disable CheckNamespace

namespace Bots.DungeonBuddy.DungeonScripts.Legion
// ReSharper restore CheckNamespace
{
    public abstract class TombOfSargeras : Dungeon
    {
        #region Overrides of Dungeon
        public override Vector3 Entrance { get; } = new Vector3(-534.6477f, 2431.087f, 108.2416f);

        public override Vector3 ExitLocation { get; } = Vector3.Zero;

        public override void OnEnter()
        {
            if (Me.IsTank())
            {
                Alert.Show(
                    "Tanking Not Supported",
                    string.Format(
                        "Tanking is not supported in the {0} script. " +
                        "If you wish to stay in raid and play manually then press 'Continue', otherwise you will automatically leave raid.",
                        Name),
                    30,
                    true,
                    true,
                    null,
                    () => Lua.DoString("LeaveParty()"),
                    "Continue",
                    "Leave");
            }
            else
            {
                Alert.Show(
                    "Do Not AFK",
                    "It is highly recommended you do not afk while in a raid and be prepared to intervene " +
                    "if needed in the event something goes wrong or you're asked to perform a certain task.",
                    20,
                    true,
                    false,
                    null,
                    null,
                    "Ok");
            }
        }

        #endregion

        protected static LocalPlayer Me => StyxWoW.Me;
    }

    public class TheGatesOfHell : TombOfSargeras
    {
        #region Overrides of Dungeon

        public override uint DungeonId => 1494;

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            units.RemoveAll(
                ret =>
                {
                    var unit = ret as WoWUnit;
                    if (unit == null)
                        return false;

                    if (unit.Entry == MobId_InfernalChaosbringer && unit.CastingSpellId == SpellId_MassiveEruption && Me.IsMelee)
                        return true;

                    if (unit.Entry == MobId_UnderseaCustodian && unit.CastingSpellId == SpellId_ElectricShock && Me.IsMelee)
                        return true;

                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            var isTank = Me.IsTank;
            var inDungeon = Me.MapId == LfgDungeon.MapId;

            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit == null)
                    continue;
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            var isDps = Me.IsDps;

            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit == null)
                    continue;

                switch(unit.Entry)
                {
                    case MobId_RazorjawWavemender:
                        priority.Score += 5000;
                        break;
                    case MobId_RazorjawGladiator:
                        priority.Score += 5500;
                        break;
                }
            }
        }

        public override void OnEnter()
        {
            base.OnEnter();
            _capabilityHandler = CapabilityManager.Instance.CreateNewHandle();
        }

        public override void OnExit()
        {
            base.OnExit();
            if (_capabilityHandler != null)
            {
                CapabilityManager.Instance.Clear(_capabilityHandler);
                _capabilityHandler = null;
            }
        }

        #endregion

        #region Root


        [EncounterHandler(0, "Root Handler")]
        public Func<WoWUnit, Task<bool>> RootBehavior()
        {
            CapabilityManagerHandle handle = CapabilityManager.Instance.CreateNewHandle();
            return async npc =>
            {
                return false;
            };
        }

        #endregion

        #region Goroth

        #region Trash

        private const uint MobId_InfernalChaosbringer = 118022;
        private const int SpellId_MassiveEruption = 242909;
        private CapabilityManagerHandle _capabilityHandler;

        [EncounterHandler(MobId_InfernalChaosbringer, "Infernal Chaosbringer")]
        public Func<WoWUnit, Task<bool>> InfernalChaosbringerHandler()
        {
            AddAvoidObject<WoWUnit>(() => true, 15, u => u.Entry == MobId_InfernalChaosbringer && u.CastingSpellId == SpellId_MassiveEruption);

            return async boss =>
            {
                if (boss.CastingSpellId == SpellId_MassiveEruption
                    && Me.CurrentTargetGuid == boss.Guid 
                        || (Me.GotTarget && Me.CurrentTarget.Location.DistanceSquared(boss.Location) < 10*10))
                {
                    CapabilityManager.Instance.Update(_capabilityHandler, CapabilityFlags.GapCloser, 5000, "Massive Eruption");
                }
                return false;
            };
        }

        #endregion

        private const uint GameObjectId_InfernalSpike = 266938;
        private const uint MobId_Goroth = 115844;

        private const int SpellId_InfernalBurning = 233062;
        private const uint AreaTriggerId_ShatteringStar = 13412;

        [EncounterHandler(MobId_Goroth, "Goroth")]
        public Func<WoWUnit, Task<bool>> GorothHandler()
        {
            WoWUnit goroth = null;
            Func<WoWUnit, WoWGameObject, Vector3> calculateSpikeLosLoc
                = (boss, spike) => WoWMathHelper.CalculatePointFrom(boss.Location, spike.Location, -5);

            PerFrameCachedValue<WoWUnit> shatteringStarTarget = new PerFrameCachedValue<WoWUnit>(
                () => ScriptHelpers.IsViable(goroth) && goroth.Combat
                    ?  Me.RaidMembers.FirstOrDefault(r => r.HasAura("Shattering Star")) : null);


            AddAvoidLine<WoWUnit>(() => shatteringStarTarget.Value != null && shatteringStarTarget.Value != Me,
                // set start point back some to cause toons to not stand too close to boss.
                u => u.Location.RayCast(WoWMathHelper.CalculateNeededFacing(shatteringStarTarget.Value.Location, u.Location), 6),
                u => WoWMathHelper.CalculateNeededFacing(u.Location, shatteringStarTarget.Value.Location),
                u => u.Location.Distance(shatteringStarTarget.Value.Location) + 12,
                u => 12,
                u => u.Entry == MobId_Goroth);

            AddAvoidObject<WoWGameObject>(4.5f, GameObjectId_InfernalSpike);

            return async boss =>
            {
                goroth = boss;
                var isShatteringStarTargetOnMe = shatteringStarTarget.Value == Me;

                if (boss.CastingSpellId == SpellId_InfernalBurning || isShatteringStarTargetOnMe)
                {
                    var spikeLosLoc = (from gObj in ObjectManager.GetObjectsOfType<WoWGameObject>()
                                       where gObj.Entry == GameObjectId_InfernalSpike
                                       let losLoc = calculateSpikeLosLoc(boss, gObj)
                                       orderby Me.Location.DistanceSquared(losLoc)
                                       select losLoc).FirstOrDefault();

                    if (spikeLosLoc != Vector3.Zero)
                    {
                        string reason = isShatteringStarTargetOnMe
                            ? "Shattering Star"
                            : "Infernal Spike LOS";

                        return await ScriptHelpers.StayAtLocationWhile(
                            () => ScriptHelpers.IsViable(boss) && boss.CastingSpellId == SpellId_InfernalBurning && spikeLosLoc != Vector3.Zero,
                            spikeLosLoc, reason, 1);
                    }
                }

                return false;
            };
        }

        #endregion

        #region Harjatan

        private const int SpellId_DrawIn = 232061;

        private const uint MobId_Harjatan = 116407;
        private const uint AreaTriggerId_DrenchingSlough = 13251;
        private const int MissileSpellId_DrenchingSlough = 233526;
        private const int SpellId_UncheckedRage = 231854;
        private const uint MobId_RazorjawWavemender = 116569;

        [EncounterHandler(MobId_Harjatan, "Harjatan")]
        public Func<WoWUnit, Task<bool>> HarjatanHandler()
        {
            WoWUnit harjatan = null;

            AddAvoidObject<WoWAreaTrigger>(() => ScriptHelpers.IsViable(harjatan) && harjatan.Combat,
                a => harjatan.CastingSpellId == SpellId_DrawIn ? 10 : 6,
                a => a.Entry == AreaTriggerId_DrenchingSlough);

            AddAvoidLocation(
                () => ScriptHelpers.IsViable(harjatan) && harjatan.Combat, 
                6, m => m.ImpactPosition, 
                () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == MissileSpellId_DrenchingSlough));

            return async boss =>
            {
                harjatan = boss;
                if (boss.CastingSpellId == SpellId_UncheckedRage)
                {
                    // Stand in front to split the damage from Unchecked Rage.
                    return await ScriptHelpers.StayAtLocationWhile(
                        () => ScriptHelpers.IsViable(boss) && boss.CastingSpellId == SpellId_UncheckedRage,
                        boss.Location.RayCast(boss.Rotation, 5));
                }
                return false;
            };
        }

        private const int SpellId_WaterySplash = 240066;

        [EncounterHandler(MobId_RazorjawWavemender, "Razorjaw Wavemender")]
        public Func<WoWUnit, Task<bool>> RazorjawWavemenderHandler()
        {
            return async npc => await ScriptHelpers.InterruptCast(npc, SpellId_WaterySplash);
        }

        private const uint MobId_RazorjawGladiator = 117596;

        #endregion

        #region Mistress Sassz'ine

        #region Trash

        private const uint MobId_UnderseaCustodian = 120463;
        private const uint AreaTriggerId_LightningStorm = 15039;
        private const int SpellId_ElectricShock = 244554;
        private const int MissileSpellId_LightningStorm = 240173;

        [EncounterHandler(MobId_UnderseaCustodian, "Undersea Custodian")]
        public Func<WoWUnit, Task<bool>> UnderseaCustodianHandler()
        {
            AddAvoidObject<WoWUnit>(() => true, 8,
                u => u.Entry == MobId_UnderseaCustodian && u.CastingSpellId == SpellId_ElectricShock);

            AddAvoidObject<WoWAreaTrigger>(7.5f, AreaTriggerId_LightningStorm);
            AddAvoidLocation(() => true, m => 7.5f, m => m.ImpactPosition,
                () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == MissileSpellId_LightningStorm));

            return boss => Task.FromResult(false);
        }

        #endregion

        private const uint MobId_MistressSasszine = 116407;
        private const uint MobId_ElectrifyingJellyfish = 115896;
        private const uint MobId_Sarukel = 116843;
        private const uint AreaTriggerId_BefoulingInk = 13369;
        private const uint AreaTriggerId_SlicingTornado = 13347;

        private const uint MobId_Piranhado = 116841;
        private const uint AreaTriggerId_CrashingWave = 15104;
        private const int SpellId_CrashingWave = 232827;

        [EncounterHandler(MobId_MistressSasszine, "Mistress Sassz'ine")]
        public Func<WoWUnit, Task<bool>> MistressSasszineHandler()
        {
            WoWUnit mistressSasszine = null;

            AddAvoidLine<WoWAreaTrigger>(() => true,
                a => WoWMathHelper.CalculatePointBehind(a.Location, a.Rotation, 5),
                a => a.Rotation, a => 20, a => 10, a => a.Entry == AreaTriggerId_SlicingTornado && a.DistanceSqr < 15*15);

            AddAvoidObject<WoWUnit>(() => !Me.HasAura("Consuming Hunger"), 8, u => u.Entry == MobId_ElectrifyingJellyfish);
            AddAvoidObject<WoWUnit>(() => !Me.HasAura("Befouling Ink"), 23, u => u.Entry == MobId_Sarukel && u.HasAura("Devouring Maw"));

            Vector2[] crashingWavePoints = { new Vector2(-10, -20), new Vector2(-10, 20), new Vector2(200, 20), new Vector2(200, -20) };

            AddAvoidPolygon(
                () => true, null, 40, a => a.Rotation, a => 1,
                a => 30, a => crashingWavePoints, a => a.Location,
                () => ObjectManager.GetObjectsOfType<WoWAreaTrigger>().Where(a => a.Entry == AreaTriggerId_CrashingWave),
                a => a.IsValid);

            //AddAvoidLine<WoWUnit>(() => true,
            //    a => WoWMathHelper.CalculatePointBehind(a.Location, a.Rotation, 40),
            //    a => a.Rotation, a => 220, a => 20, a => a.Entry == MobId_Piranhado && a.CastingSpellId == SpellId_CrashingWave);

            var debugTimer = WaitTimer.FiveSeconds;
            return async boss =>
            {
                mistressSasszine = boss;

                CapabilityManager.Instance.Update(_capabilityHandler, CapabilityFlags.GapCloser,
                    () => ScriptHelpers.IsViable(boss) && boss.Combat, 
                    "Too many avoids at Mistress Sassz'ine encounter");
                
                if (debugTimer.IsFinished)
                {
                    var crashingWave = ObjectManager.GetObjectsOfType<WoWAreaTrigger>().FirstOrDefault(a => a.Entry == AreaTriggerId_CrashingWave);
                    if (crashingWave != null)
                    {
                        StyxWoW.Overlay.AddToast("Crashing Wave Active", 4000);
                        debugTimer.Reset();
                    }
                }

                if (Me.HasAura("Consuming Hunger"))
                {
                    var jellyfish = ObjectManager.GetObjectsOfType<WoWUnit>()
                        .Where(u => u.Entry == MobId_ElectrifyingJellyfish)
                        .MinByOrDefault(u => u.DistanceSqr);
                    if (jellyfish != null)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => ScriptHelpers.IsViable(jellyfish) && Me.HasAura("Consuming Hunger"), 
                            jellyfish.Location, jellyfish.SafeName, 5);
                    }
                }

                var sarukel = ObjectManager.GetObjectsOfType<WoWUnit>().FirstOrDefault(u => u.Entry == MobId_Sarukel);
                if (sarukel != null)
                {
                    if (Me.HasAura("Befouling Ink"))
                    {
                        var moveTo = WoWMathHelper.CalculatePointFrom(Me.Location, sarukel.Location, 18);
                        return (await CommonCoroutines.MoveTo(moveTo, sarukel.SafeName)).IsSuccessful();
                    }
                    var ink = ObjectManager.GetObjectsOfType<WoWAreaTrigger>()
                        .Where(a => a.Entry == AreaTriggerId_BefoulingInk)
                        .MinByOrDefault(a => a.DistanceSqr);

                    if (ink != null)
                        return (await CommonCoroutines.MoveTo(ink.Location, ink.SafeName)).IsSuccessful();
                }
      
                return false;
            };
        }
        #endregion
    }
}