﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Threading.Tasks;
using Bots.DungeonBuddy.Attributes;
using Bots.DungeonBuddy.Enums;
using Bots.DungeonBuddy.Helpers;
using Bots.DungeonBuddy.Profiles;
using Buddy.Coroutines;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Bars;
using Styx.CommonBot.Coroutines;
using Styx.CommonBot.Routines;
using Styx.Helpers;
using Styx.Pathing;
using Styx.Pathing.Avoidance;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using Styx.CommonBot.POI;

// ReSharper disable CheckNamespace

namespace Bots.DungeonBuddy.DungeonScripts.Legion
// ReSharper restore CheckNamespace
{
    public class TormentedGuardians : TheEmeraldNightmare
    {
        #region Overrides of Dungeon

        public override uint DungeonId => 1288;

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            units.RemoveAll(
                ret =>
                {
                    var unit = ret as WoWUnit;
                    if (unit == null)
                        return false;

                    if (HasHighStacksOfDragonMarkAura(unit))
                        return true;

                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            var isTank = Me.IsTank;
            var inDungeon = Me.MapId == LfgDungeon.MapId;
            var isRangedDps = Me.IsRangedDps;

            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit == null)
                    continue;

                if (isRangedDps && unit.Entry == MobId_EntanglingRoots)
                    outgoingunits.Add(unit);
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            var isDps = Me.IsDps;
            var isRangedDps = isDps && Me.IsRanged;

            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit == null)
                    continue;

                if (unit.Combat && unit.HasAura("Aura of Dread Thorns"))
                {
                    priority.Score -= 5000;
                    continue;
                }

                switch (unit.Entry)
                {
                    case MobId_NightmareAncient:
                        if (isDps)
                            priority.Score += 4500;
                        break;
                    case MobId_DefiledDruidSpirit:
                    case MobId_RottenDrake:
                        if (isDps)
                            priority.Score += 5000;
                        break;
                    case MobId_SpiritShade:
                    case MobId_TwistedSister:
                        if (isDps)
                            priority.Score += 5500;
                        break;
                    case MobId_CorruptedWisp:
                    case MobId_EntanglingRoots:
                        if (isRangedDps)
                            priority.Score += 6000;
                        break;
                }
            }
        }

        public override void IncludeLootTargetsFilter(List<WoWObject> incomingObjects, HashSet<WoWObject> outgoingObjects)
        {
            foreach (var lootTarget in incomingObjects)
            {
                var gObj = lootTarget as WoWGameObject;
                if (gObj != null && gObj.Entry == GameObjectId_SpoilsoftheNightmare && gObj.CanLoot &&
                    DungeonBuddySettings.Instance.LootMode != LootMode.Off)
                {
                    outgoingObjects.Add(gObj);
                }
            }
        }

        public override async Task<bool> HandleMovement(MoveToParameters moveToParameters)
        {
            var myLoc = (WoWMovement.ActiveMover ?? Me).Location;
            var dest = moveToParameters.Location;
            var areaAtMyLoc = _darkboughAreas.FirstOrDefault(a => a.IsInArea(myLoc));
            var areaAtDest = _darkboughAreas.FirstOrDefault(a => a.IsInArea(dest));

            if (areaAtMyLoc != null && areaAtDest != null && areaAtMyLoc != areaAtDest)
            {
                if (areaAtMyLoc.GetType() == typeof(MainArea))
                    return await areaAtDest.HandlePortIn();

                return (await CommonCoroutines.MoveTo(areaAtMyLoc.PortOutLocation)).IsSuccessful();
            }

            return false;
        }

        private DarkboughArea[] _darkboughAreas = new DarkboughArea[]
        {
                new MainArea(),
                new UrsocArea(),
                new DragonsOfNightmareArea(),
                new CenariusArea()
        };

        private class UrsocArea : DarkboughArea
        {
            public override Vector3 PortInLocation { get; } = new Vector3(1981.738f, 1343.164f, 573.0516f);
            public override Vector3 PortOutLocation { get; } = new Vector3(-12350.65f, -12954.52f, 331.2879f);
            public override bool IsInArea(Vector3 point) => point.X > -12629.41f && point.X < -12032f && point.Y > -13440.78 && point.Y < -12805.25;
        }

        private class DragonsOfNightmareArea : DarkboughArea
        {
            public override Vector3 PortInLocation { get; } = new Vector3(1922.327f, 1282.31f, 617.3365f);
            public override Vector3 PortOutLocation { get; } = new Vector3(454.324f, -12729.54f, 30.3518f);
            public override bool IsInArea(Vector3 point) => point.X > 0 && point.X < 1595.176f && point.Y < -12297.12f;
        }

        private class CenariusArea : DarkboughArea
        {
            public override Vector3 PortInLocation { get; } = new Vector3(1876.323f, 1325.014f, 597.8997f);
            public override Vector3 PortOutLocation { get; } = new Vector3(11180.23f, -12622.58f, 553.2471f);
            public override bool IsInArea(Vector3 point) => point.X > 10530.67 && point.Y < -11686.73;

            public override async Task<bool> HandlePortIn()
            {
                var nightmareIris = ObjectManager.GetObjectsOfType<WoWGameObject>().FirstOrDefault(g => g.Entry == 251980);
                if (nightmareIris != null && nightmareIris.Distance < 8*8)
                {
                    nightmareIris.Interact();
                    await Coroutine.Sleep(4000);
                    return true;
                }
                return await base.HandlePortIn();
            }
        }

        #endregion

        #region Root


        [EncounterHandler(0, "Root Handler")]
        public Func<WoWUnit, Task<bool>> RootBehavior()
        {
            CapabilityManagerHandle handle = CapabilityManager.Instance.CreateNewHandle();
            return async npc =>
            {
                return false;
            };
        }

        #endregion

        #region Ursoc

        private const uint MobId_Ursoc = 100497;

        [EncounterHandler(MobId_Ursoc, "Ursoc")]
        public Func<WoWUnit, Task<bool>> UrsocHandler()
        {
            TimeCachedValue<Vector3> groupLoc = new TimeCachedValue<Vector3>(3000, 
                () => ScriptHelpers.GetGroupCenterLocation(p => !p.IsTank && p.IsAlive, 15));

            PerFrameCachedValue<WoWUnit> focusedGazeTarget = new PerFrameCachedValue<WoWUnit>(() => Me.RaidMembers.FirstOrDefault(r => r.HasAura("Focused Gaze")));

            AddAvoidLine<WoWUnit>(() => Me.HasAura("Momentum") && focusedGazeTarget.Value != null,
                u => u.Location, 
                u => WoWMathHelper.CalculateNeededFacing(u.Location, focusedGazeTarget.Value.Location), 
                u => focusedGazeTarget.Value.Location.Distance(u.Location), 
                u => 10, u => u.Entry == MobId_Ursoc && u.CastingSpellId == 232);

            return async boss =>
            {
                if (focusedGazeTarget.Value != null)
                {
                    Vector3 moveTo;
                    if (focusedGazeTarget.Value.IsMe)
                    {
                        moveTo = WoWMathHelper.CalculatePointFrom(boss.Location, groupLoc, -15);
                        return await ScriptHelpers.StayAtLocationWhile(() => Me.HasAura("Focused Gaze"), moveTo, "Focused Gaze Target", 4);
                    }
                    if (!Me.HasAura("Momentum"))
                    {
                        var fromLoc = boss.Location.RayCast(WoWMathHelper.CalculateNeededFacing(boss.Location, focusedGazeTarget.Value.Location), 3);
                        var toLoc = WoWMathHelper.CalculatePointFrom(boss.Location, focusedGazeTarget.Value.Location, 3);
                        moveTo = Me.Location.GetNearestPointOnSegment(fromLoc, toLoc);
                        return await ScriptHelpers.StayAtLocationWhile(() => ScriptHelpers.IsViable(focusedGazeTarget), moveTo, 
                            "Focused Gaze Intercept", 2);
                    }
                }

                return false;
            };
        }

        #endregion

        #region Dragons of Nightmare

        private HashSet<uint> _dragonsOfNightmare = new HashSet<uint>()
        {
            MobId_Ysondre,MobId_Emeriss,MobId_Lethon,MobId_Taerar
        };

        private bool HasHighStacksOfDragonMarkAura(WoWUnit unit)
        {
            var auraName = GetDragonMarkName(unit);
            if (auraName == null)
                return false;

            var aura = Me.GetAuraByName(auraName);
            return aura != null && aura.StackCount >= 8;
        }

        private string GetDragonMarkName(WoWUnit unit)
        {
            switch (unit.Entry)
            {
                case MobId_Ysondre:
                    return "Mark of Ysondre";
                case MobId_Emeriss:
                    return "Mark of Emeriss";
                case MobId_Lethon:
                    return "Mark of Lethon";
                case MobId_Taerar:
                    return "Mark of Taerar";
                default:
                    return null;
            }
        }

        private const uint MobId_Ysondre = 102679;
        private const int SpellId_CorruptedBreath = 203028;
        private const uint MobId_NightmareBloom = 102804;
        private const uint MobId_DefiledDruidSpirit = 103080;

        [EncounterHandler(MobId_Ysondre, "Ysondre")]
        public Func<WoWUnit, Task<bool>> YsondreHandler()
        {
            AddAvoidObject<WoWUnit>(() => true, 45, u => HasHighStacksOfDragonMarkAura(u));
            AddAvoidUnitCone(0, 45, 35, u => u.Combat && u.CastingSpellId == SpellId_CorruptedBreath);

            return async boss =>
            {
                if (Me.IsRangedDps)
                {
                    WoWUnit nightmareBloom = null;
                    var nightmareBloomAura = Me.GetAuraByName("Nightmare Bloom");
                    if (nightmareBloomAura != null)
                    {
                        nightmareBloom = ObjectManager.GetObjectByGuid<WoWUnit>(nightmareBloomAura.CreatorGuid);
                        Func<bool> stayAtNightmareBloom = () => ScriptHelpers.IsViable(nightmareBloom)
                            && !Me.RaidMembers.Any(r => !r.IsMe && r.HasAura("Nightmare Bloom", a => a.CreatorGuid == nightmareBloom.Guid));

                        if (stayAtNightmareBloom())
                            return await ScriptHelpers.StayAtLocationWhile(stayAtNightmareBloom, nightmareBloom.Location, "Nightmare Bloom", 4);
                    }

                    nightmareBloom = ObjectManager.GetObjectsOfType<WoWUnit>()
                        .Where(u => u.Entry == MobId_NightmareBloom)
                        .OrderBy(u => u.DistanceSqr)
                        .Where(u => !Me.RaidMembers.Any(r => r.Location.DistanceSquared(u.Location) < 5 * 5))
                        .FirstOrDefault();

                    if (nightmareBloom != null)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => ScriptHelpers.IsViable(nightmareBloom), nightmareBloom.Location, "Nightmare Bloom", 4);
                    }
                }

                if (await ScriptHelpers.DispelGroup("Defiled Vines", ScriptHelpers.PartyDispelType.Magic))
                    return true;

                return false;
            };
        }

        private const uint MobId_Emeriss = 102683;

        [EncounterHandler(MobId_Emeriss, "Emeriss")]
        public Func<WoWUnit, Task<bool>> EmerissHandler()
        {
            return async boss =>
            {
                return false;
            };
        }

        private const uint MobId_Lethon = 102682;
        private const uint MobId_SpiritShade = 103100;

        private const int MissileSpellId_Gloom = 205866;
        [EncounterHandler(MobId_Lethon, "Lethon")]
        public Func<WoWUnit, Task<bool>> LethonHandler()
        {
            AddAvoidLocation(() => true, 10, m => m.ImpactPosition, 
                () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == MissileSpellId_Gloom));

            return async boss =>
            {
                return false;
            };
        }

        private const uint MobId_Taerar = 102681;
        private const uint AreaTriggerId_SeepingFog = 10791;

        [EncounterHandler(MobId_Taerar, "Taerar")]
        public Func<WoWUnit, Task<bool>> TaerarHandler()
        {
            AddAvoidObject<WoWAreaTrigger>(5, AreaTriggerId_SeepingFog);
            return async boss =>
            {
                return await ScriptHelpers.DispelGroup("Seeping Fog", ScriptHelpers.PartyDispelType.Magic);
            };
        }

        #endregion

        #region Cenarius

        #region Trash
        private const uint MobId_GuardianGorroc = 111856;

        private const uint MobId_DreadsoulCorruptor = 112261;
        private const int SpellId_ShadowVolley = 222939;

        [EncounterHandler(MobId_GuardianGorroc, "Guardian Gorroc")]
        [EncounterHandler(MobId_DreadsoulCorruptor, "Dreadsoul Corruptor")]
        public Func<WoWUnit, Task<bool>> DreadsoulCorruptorHandler()
        {
            return async boss => await ScriptHelpers.InterruptCast(boss, SpellId_ShadowVolley);
        }

        private const uint AreaTriggerId_Darkfall = 12516;
        private const int SpellId_EruptingTerror = 223038;
        private const int SpellId_Darkfall = 223590;


        private const uint MobId_DreadsoulDefiler = 112260;
        [EncounterHandler(MobId_DreadsoulDefiler, "Dreadsoul Defiler")]
        public Func<WoWUnit, Task<bool>> DreadsoulDefilerHandler()
        {
            AddAvoidObject<WoWAreaTrigger>(12, AreaTriggerId_Darkfall);

            return async boss => await ScriptHelpers.InterruptCast(boss, SpellId_EruptingTerror, SpellId_Darkfall);
        }


        private const int MissileSpellId_HauntingBurst = 223647;
        private const uint MobId_HauntingNightmare = 112320;
        [EncounterHandler(MobId_HauntingNightmare, "Haunting Nightmare")]
        public Func<WoWUnit, Task<bool>> HauntingNightmareHandler()
        {
            PerFrameCachedValue<WoWMissile> hauntingBurstMissile = new PerFrameCachedValue<WoWMissile>(
                () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == MissileSpellId_HauntingBurst)
                    .MinByOrDefault(m => m.ImpactPosition.DistanceSquared(Me.Location)));

            return async boss =>
            {
                if (hauntingBurstMissile.Value != null)
                {
                    return await ScriptHelpers.StayAtLocationWhile(
                        () => hauntingBurstMissile.Value != null, 
                        hauntingBurstMissile.Value.ImpactPosition, 
                        hauntingBurstMissile.Value.Spell.Name, 
                        4);
                }

                return false;
            };
        }
        #endregion


        private const uint MobId_CorruptedWisp= 106304;
        private const uint MobId_TwistedSister = 105495;
        private const uint MobId_RottenDrake = 105494;
        private const int SpellId_RottenBreath = 211192;
        private const uint AreaTriggerId_CleansedGround = 11427;
        private const uint AreaTriggerId_NightmareBrambles = 11310;
        private const uint MobId_Cenarius = 104636;
        private const uint MobId_EntanglingRoots = 108040;

        private const uint GameObjectId_SpoilsoftheNightmare = 254168;

        [EncounterHandler(MobId_Cenarius, "Cenarius")]
        public Func<WoWUnit, Task<bool>> CenariusHandler()
        {
            AddAvoidObject<WoWUnit>(() => true, 12, u => u.Entry == MobId_CorruptedWisp && u.CurrentTargetGuid == Me.Guid);
            AddAvoidUnitCone(0, 45, 60, 
                u => u.Entry == MobId_RottenDrake && u.CastingSpellId == SpellId_RottenBreath);

            AddAvoidObject<WoWAreaTrigger>(() => !Me.HasAura("Unbound Touch") && !Me.HasAura("Unbound Essence"), 4, a => a.Entry == AreaTriggerId_NightmareBrambles);

            AddAvoidObject<WoWAreaTrigger>(
                () => Me.HasAura("Creeping Nightmares", a => a.StackCount < 17),
                12,
                a => a.Entry == AreaTriggerId_CleansedGround);

            return async boss =>
            {
                var unboundTouchAura = Me.GetAuraByName("Unbound Touch");
                var unboundEssenceAura = Me.GetAuraByName("Unbound Essence");

                if (unboundTouchAura != null || unboundEssenceAura != null)
                {
                    if (unboundTouchAura == null || unboundTouchAura.TimeLeft > TimeSpan.FromSeconds(4))
                    {
                        var nightmareBrambles = ObjectManager.GetObjectsOfType<WoWAreaTrigger>().Where(a => a.Entry == AreaTriggerId_NightmareBrambles)
                            .MinByOrDefault(a => a.DistanceSqr);

                        if (nightmareBrambles != null)
                            return (await CommonCoroutines.MoveTo(nightmareBrambles.Location, nightmareBrambles.SafeName)).IsSuccessful();
                    }

                    if (unboundTouchAura != null)
                    {
                        var nearestPlayer = Me.RaidMembers.Where(p => !p.IsMe && p.IsAlive && p.IsDps).MinByOrDefault(p => p.DistanceSqr);
                        if (nearestPlayer != null)
                        {
                            return await ScriptHelpers.StayAtLocationWhile(
                                () => Me.HasAura("Unbound Touch"),
                                nearestPlayer.Location,
                                "Unbound Touch Spread location",
                                3);
                        }
                    }
                }

                if (Me.HasAura("Creeping Nightmares", a => a.StackCount >= 17))
                {
                    var cleansedGround = ObjectManager.GetObjectsOfType<WoWAreaTrigger>()
                        .Where(a => a.Entry == AreaTriggerId_CleansedGround)
                        .MinByOrDefault(a => a.DistanceSqr);
                    if (cleansedGround != null)
                        return (await CommonCoroutines.MoveTo(cleansedGround.Location, cleansedGround.SafeName)).IsSuccessful();
                }


                if (await ScriptHelpers.DispelEnemy("Twisted Touch of Life", ScriptHelpers.EnemyDispelType.Magic, Me.CurrentTarget))
                    return true;

                if (await ScriptHelpers.DispelGroup("Nightmare Brambles", ScriptHelpers.PartyDispelType.Magic))
                    return true;

                return false;
            };
        }

        private const int SpellId_TwistedTouchofLife = 211368;
        private const int MissileSpellId_NightmareJavelin = 211498;

        [EncounterHandler(MobId_TwistedSister, "Twisted Sister")]
        public Func<WoWUnit, Task<bool>> TwistedSisterHandler()
        {
            AddAvoidObject<WoWUnit>(() => true, 12, u => u.Entry == MobId_CorruptedWisp && u.CurrentTargetGuid == Me.Guid);
            AddAvoidLocation(() => true, 4, m => m.ImpactPosition,
                () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == MissileSpellId_NightmareJavelin));

            AddAvoidObject<WoWPlayer>(() => !Me.HasAura("Scorned Touch"), 8, p => !p.IsMe && p.IsAlive && (p.HasAura("Scorned Touch") || Me.HasAura("Scorned Touch")));

            return async boss =>
            {
                return await ScriptHelpers.InterruptCast(boss, SpellId_TwistedTouchofLife);
            };
        }

        private const int SpellId_DesiccatingStomp = 211073;

        private const uint MobId_NightmareAncient = 105468;
        [EncounterHandler(MobId_NightmareAncient, "Nightmare Ancient")]
        public Func<WoWUnit, Task<bool>> NightmareAncientHandler()
        {
            return async boss =>
            {
                return boss.IsHostile &&
                    await ScriptHelpers.StayAtLocationWhile(
                        () => ScriptHelpers.IsViable(boss) && boss.CastingSpellId == SpellId_DesiccatingStomp,
                        boss.Location, boss.SafeName, 7);
            };
        }

        #endregion
    }

}