﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Threading.Tasks;
using Bots.DungeonBuddy.Attributes;
using Bots.DungeonBuddy.Enums;
using Bots.DungeonBuddy.Helpers;
using Bots.DungeonBuddy.Profiles;
using Buddy.Coroutines;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.Common.Helpers;
using Styx.CommonBot;
using Styx.CommonBot.Bars;
using Styx.CommonBot.Coroutines;
using Styx.CommonBot.Routines;
using Styx.Helpers;
using Styx.Pathing;
using Styx.Pathing.Avoidance;
using Styx.TreeSharp;
using Styx.WoWInternals;
using Styx.WoWInternals.WoWObjects;
using System.Diagnostics;
using System.Runtime.CompilerServices;

// ReSharper disable CheckNamespace

namespace Bots.DungeonBuddy.DungeonScripts.Legion
// ReSharper restore CheckNamespace
{
    public class WailingHalls : TombOfSargeras
    {
        #region Overrides of Dungeon

        public override uint DungeonId => 1495;

        public override void RemoveTargetsFilter(List<WoWObject> units)
        {
            units.RemoveAll(
                ret =>
                {
                    var unit = ret as WoWUnit;
                    if (unit == null)
                        return false;

                    if (unit.Entry == MobId_Atrigan && unit.HasAura("Bone Saw"))
                        return true;

                    if (unit.Entry == MobId_Belac && unit.HasAura("Fel Squall"))
                        return true;

                    // These explode upon death 
                    if (unit.Entry == MobId_WailingSoul && unit.HealthPercent < 10 && Me.IsMelee)
                        return true;

                    return false;
                });
        }

        public override void IncludeTargetsFilter(List<WoWObject> incomingunits, HashSet<WoWObject> outgoingunits)
        {
            var isTank = Me.IsTank;
            var inDungeon = Me.MapId == LfgDungeon.MapId;

            foreach (var obj in incomingunits)
            {
                var unit = obj as WoWUnit;
                if (unit == null)
                    continue;
            }
        }

        public override void WeighTargetsFilter(List<Targeting.TargetPriority> units)
        {
            var isDps = Me.IsDps;

            foreach (var priority in units)
            {
                var unit = priority.Object as WoWUnit;
                if (unit == null)
                    continue;

                switch (unit.Entry)
                {
                    case MobId_Moontalon:
                    case MobId_GhastlyBonewarden:
                        priority.Score += 5000;
                        break;
                    case MobId_ReanimatedTemplar:
                        priority.Score += 5000;
                        break;
                    case MobId_TormentedSoul:
                        if (unit.DistanceSqr < 60*60 && unit.ZDiff < 10)
                            priority.Score += 5000;
                        break;
                }
            }
        }

        private CapabilityManagerHandle _capabilityHandler;

        public override void OnEnter()
        {
            base.OnEnter();
            _capabilityHandler = CapabilityManager.Instance.CreateNewHandle();
        }

        public override void OnExit()
        {
            base.OnExit();
            if (_capabilityHandler != null)
            {
                CapabilityManager.Instance.Clear(_capabilityHandler);
                _capabilityHandler = null;
            }
        }

        #endregion

        #region Root

        [EncounterHandler(0, "Root Handler")]
        public Func<WoWUnit, Task<bool>> RootBehavior()
        {
            return async npc =>
            {
                return false;
            };
        }

        #endregion

        #region Demonic Inquisition

        protected const uint MobId_Belac = 116691;
        private const int SpellId_PangsofGuilt = 239401;
        private const uint AreaTriggerId_RemnantofHope = 15205;

        private const uint MobId_TormentedSoul = 117957;

        [EncounterHandler(MobId_Belac, "Belac", Mode =CallBehaviorMode.Proximity, BossRange = 100)]
        public Func<WoWUnit, Task<bool>> BelacBehavior()
        {
            AddAvoidObject<WoWPlayer>(() => true, 8, p => !p.IsMe && (p.HasAura("Echoing Anguish") || Me.HasAura("Echoing Anguish")));
            AddAvoidObject<WoWUnit>(() => true, 12, u => u.Entry == MobId_Belac && u.HasAura("Fel Squall"));
            var leftDoorLoc = new Vector3(6344.998f, -788.6657f, 3013.384f);
            var rightDoorLoc = new Vector3(6345.314f, -802.8139f, 3013.655f);
            var randomPointInsideRoom = WoWMathHelper.GetRandomPointInCircle(new Vector3(6350.645f, -795.6928f, 3013.117f), 5);
            var groupLoc = new TimeCachedValue<Vector3>(5000, () => ScriptHelpers.GetGroupCenterLocation());
            var cageCenter = new Vector3(6403.36f, -933.2507f, 3037.168f);

            return async boss =>
            {
                if (await ScriptHelpers.MoveInsideBossRoom(boss, leftDoorLoc, rightDoorLoc, randomPointInsideRoom))
                    return true;

                if (!boss.Combat)
                    return false;

                CapabilityManager.Instance.Update(_capabilityHandler, CapabilityFlags.GapCloser,
                    () => ScriptHelpers.IsViable(boss) && boss.Combat, 
                    "Prevents gap-closing into a deadly effect such as Bone Saw");

                if (Me.HasAura("Unbearable Torment") && ActionBar.Extra.IsActive && ActionBar.Extra.Buttons[0].CanUse)
                {
                    ActionBar.Extra.Buttons[0].Use();
                    return true;
                }

                return await ScriptHelpers.InterruptCast(boss, SpellId_PangsofGuilt);
            };
        }

        protected const uint MobId_Atrigan = 116689;
        private const int SpellId_ScytheSweep = 233426;

        [EncounterHandler(MobId_Atrigan, "Atrigan")]
        public Func<WoWUnit, Task<bool>> AtriganBehavior()
        {
            AddAvoidUnitCone(0, 15, 120, u => u.Entry == MobId_Atrigan && u.CastingSpellId == SpellId_ScytheSweep);
            AddAvoidObject<WoWUnit>(() => true, 12, u => u.Entry == MobId_Atrigan && u.HasAura("Bone Saw"));

            return async boss =>
            {
                return false;
            };
        }

        private const int MissileSpellId_TormentedEruption = 239136;
        private const int SpellId_RemnantOfHope = 235293;

        [EncounterHandler(MobId_TormentedSoul, "Tormented Soul", Mode = CallBehaviorMode.Proximity)]
        public Func<WoWUnit, Task<bool>> TormentedSoulBehavior()
        {
            AddAvoidLocation(() => true, 3, m => m.ImpactPosition,
                () => WoWMissile.InFlightMissiles.Where(m => m.SpellId == MissileSpellId_TormentedEruption));

            return async boss =>
            {
                if (ActionBar.Extra.IsActive && ActionBar.Extra.Buttons[0].Id == SpellId_RemnantOfHope)
                {
                    ActionBar.Extra.Buttons[0].Use();
                    return true;
                }

                var reminantsOfHope = ObjectManager.GetObjectsOfType<WoWAreaTrigger>().Where(a => a.Entry == AreaTriggerId_RemnantofHope)
                    .MinByOrDefault(a => a.DistanceSqr);

                if (reminantsOfHope != null)
                    return (await CommonCoroutines.MoveTo(reminantsOfHope.Location, reminantsOfHope.SafeName)).IsSuccessful();

                return false;
            };
        }

        #endregion

        #region Sisters of the Moon

        private const uint AreaTriggerId_LunarBarrage = 14722;
        protected const uint MobId_HuntressKasparian = 118523;
        protected const uint MobId_CaptainYathaeMoonstrike = 118374;
        protected const uint MobId_PriestessLunaspyre = 118518;
        private const uint AreaTriggerId_MoonAreaTriggerDarkside = 13568;
        private const uint AreaTriggerId_MoonAreaTrigger = 13559;
        private const uint MobId_Moontalon = 119205;

        [EncounterHandler(MobId_CaptainYathaeMoonstrike, "Captain Yathae Moonstrike", Mode = CallBehaviorMode.Proximity, BossRange = 100)]
        public Func<WoWUnit, Task<bool>> CaptainYathaeMoonstrikeBehavior()
        {
            AddAvoidObject<WoWAreaTrigger>(8, AreaTriggerId_LunarBarrage);

            PerFrameCachedValue<WoWAreaTrigger> lightMoonArea =
                new PerFrameCachedValue<WoWAreaTrigger>(
                    () => ObjectManager.GetObjectsOfType<WoWAreaTrigger>().FirstOrDefault(a => a.Entry == AreaTriggerId_MoonAreaTrigger));

            PerFrameCachedValue<WoWAreaTrigger> darkMoonArea =
                new PerFrameCachedValue<WoWAreaTrigger>(
                    () => ObjectManager.GetObjectsOfType<WoWAreaTrigger>().FirstOrDefault(a => a.Entry == AreaTriggerId_MoonAreaTriggerDarkside));

            Func<bool> clearUmbraSuffusion = 
                () => Me.HasAura("Umbra Suffusion", a => a.StackCount > 10 || Me.HasAura("Rapid Shot")) && lightMoonArea.Value != null;

            Func<bool> clearLunarSuffusion = 
                () => Me.HasAura("Lunar Suffusio", a => a.StackCount > 10 || Me.HasAura("Moon Burn")) && darkMoonArea.Value != null;

            var leftDoorLoc = new Vector3(6168.022f, -1004.313f, 2899.466f);
            var rightDoorLoc = new Vector3(6163.057f, -1017.944f, 2899.466f);
            var randomPointInsideRoom = WoWMathHelper.GetRandomPointInCircle(new Vector3(6169.319f, -1013.337f, 2899.473f), 3);
            var groupLoc = new TimeCachedValue<Vector3>(5000, () => ScriptHelpers.GetGroupCenterLocation());

            return async boss =>
            {
                if (await ScriptHelpers.MoveInsideBossRoom(boss, leftDoorLoc, rightDoorLoc, randomPointInsideRoom))
                    return true;

                if (!boss.Combat)
                    return false;

                if (clearUmbraSuffusion())
                {
                    return await ScriptHelpers.StayAtLocationWhile(
                        clearUmbraSuffusion,
                        lightMoonArea.Value.Location, "Light side of Moon", 14);
                }

                if (clearLunarSuffusion())
                {
                    return await ScriptHelpers.StayAtLocationWhile(
                        clearLunarSuffusion,
                        darkMoonArea.Value.Location, "Dark side of Moon", 14);
                }

                return false;
            };
        }

        [EncounterHandler(MobId_PriestessLunaspyre, "Priestess Lunaspyre")]
        public Func<WoWUnit, Task<bool>> PriestessLunaspyreBehavior()
        {
            AddAvoidObject<WoWPlayer>(() => true, 4, p => !p.IsMe 
                && (p.HasAura("Embrace of the Eclipse", a => a.TimeLeft <= TimeSpan.FromSeconds(2)) 
                    || Me.HasAura("Embrace of the Eclipse", a => a.TimeLeft <= TimeSpan.FromSeconds(2))));

            return async boss =>
            {
                return false;
            };
        }

        #endregion

        #region The Desolate Host

        #region Trash

        private const uint MobId_UmbralGuard = 122850;
        [EncounterHandler(MobId_UmbralGuard, "Umbral Guard")]
        public Func<WoWUnit, Task<bool>> UmbralGuardBehavior()
        {
            AddAvoidUnitCone(0, 15, 45, u => u.Entry == MobId_UmbralGuard && u.Combat);

            return boss => Task.FromResult(false);
        }

        private const uint MobId_WailingSoul = 120188;
        [EncounterHandler(MobId_WailingSoul, "Wailing Soul")]
        public Func<WoWUnit, Task<bool>> WailingSoulBehavior()
        {
            // Commented out because it might not be needed.
            // AddAvoidObject<WoWUnit>(() => true, 10, u => u.Entry == MobId_WailingSoul && u.HealthPercent < 10);

            return boss => Task.FromResult(false);
        }

        #endregion

        protected const uint MobId_EngineOfSouls = 118460;
        private const uint AreaTriggerId_CollapsingFissure = 13657;
        private const uint AreaTriggerId_SoulRot = 14675;
        private const uint AreaTriggerId_TormentedCries = 13676;

        [EncounterHandler(MobId_EngineOfSouls, "Engine of Souls", Mode = CallBehaviorMode.Proximity, BossRange = 120)]
        public Func<WoWUnit, Task<bool>> EngineOfSoulsBehavior()
        {
            AddAvoidObject<WoWPlayer>(() => !Me.HasAura("Soulbind"), 10, p => p.HasAura("Soulbind"));
            AddAvoidObject<WoWAreaTrigger>(6, AreaTriggerId_CollapsingFissure);
            AddAvoidObject<WoWAreaTrigger>(2, AreaTriggerId_SoulRot);

            Vector2[] tormentedCriesPoints = {new Vector2(0, 1), new Vector2(0, -1), new Vector2(100, -10), new Vector2(100, 10) };
            AddAvoidPolygon(() => true, null, 40, a => a.Rotation, a => 1, a => 30,
                a => tormentedCriesPoints, a => a.Location, 
                () => ObjectManager.GetObjectsOfType<WoWAreaTrigger>().Where(a => a.Entry == AreaTriggerId_TormentedCries),
                u => u.IsValid, 
                true);


            var leftDoorLoc = new Vector3(6405.047f, -1056.001f, 2881.004f);
            var rightDoorLoc = new Vector3(6395.304f, -1070.434f, 2881.019f);
            var randomPointInsideRoom = WoWMathHelper.GetRandomPointInCircle(new Vector3(6406.676f, -1066.82f, 2881.826f), 3);
            var groupLoc = new TimeCachedValue<Vector3>(5000, () => ScriptHelpers.GetGroupCenterLocation());

            return async boss =>
            {
                if (await ScriptHelpers.MoveInsideBossRoom(boss, leftDoorLoc, rightDoorLoc, randomPointInsideRoom))
                    return true;

                if (!boss.Combat)
                    return false;

                if (Me.HasAura("Soulbind"))
                {
                    var soulbindPartner = ObjectManager.GetObjectsOfType<WoWPlayer>().FirstOrDefault(p => !p.IsMe && p.HasAura("Soulbind"));
                    if (soulbindPartner != null)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => Me.HasAura("Soulbind") && soulbindPartner.IsValid,
                            soulbindPartner.Location, "Soulbind meet", 4);
                    }
                }

                if (Me.HasAura("Shattering Scream"))
                {
                    var target = ObjectManager.GetObjectsOfType<WoWUnit>()
                        .Where(u => (u.Entry == MobId_GhastlyBonewarden || u.Entry == MobId_GhastlyBonewarden) && u.HasAura("Bonecage Armor"))
                        .MinByOrDefault(u => u.DistanceSqr);

                    if (target != null)
                    {
                        return await ScriptHelpers.StayAtLocationWhile(
                            () => ScriptHelpers.IsViable(target) && target.HasAura("Bonecage Armor") && Me.HasAura("Shattering Scream"),
                            target.Location, target.SafeName, 6);
                    }
                }
                return false;
            };
        }

        private const uint MobId_ReanimatedTemplar = 118715;
        private const uint AreaTriggerId_RupturingSlam = 13674;

        [EncounterHandler(MobId_ReanimatedTemplar, "Reanimated Templar")]
        public Func<WoWUnit, Task<bool>> ReanimatedTemplarBehavior()
        {
            AddAvoidObject<WoWAreaTrigger>(6, AreaTriggerId_RupturingSlam);
            return boss => Task.FromResult(false);
        }

        private const uint MobId_GhastlyBonewarden = 118728;

        #endregion

    }
}